/*-
 * Copyright (c) 2008 Jean-Yves Moulin <jym@ldh.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <getopt.h>
#include <math.h>
#include <kvm.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <sys/malloc.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <curses.h>
#ifdef __FreeBSD__
#include <libutil.h>
#include <memstat.h>
#endif
#ifdef __NetBSD__
#include <sys/mallocvar.h>
#endif

#include "lbdr.h"
#include "common.h"


extern char *optarg;
extern int optind;
extern int optopt;
extern int opterr;
extern int optreset;
extern int errno;

#ifdef __NetBSD__
kvm_t *kd;
struct nlist namelist[] =
{
#define	X_BOOTTIME	0
	{ .n_name = "_boottime" },
#define	X_HZ		1
	{ .n_name = "_hz" },
#define	X_STATHZ	2
	{ .n_name = "_stathz" },
#define	X_NCHSTATS	3
	{ .n_name = "_nchstats" },
#define	X_KMEMSTAT	4
	{ .n_name = "_kmemstatistics" },
#define	X_KMEMBUCKETS	5
	{ .n_name = "_kmembuckets" },
#define	X_ALLEVENTS	6
	{ .n_name = "_allevents" },
#define	X_POOLHEAD	7
	{ .n_name = "_pool_head" },
#define	X_UVMEXP	8
	{ .n_name = "_uvmexp" },
#define	X_TIME_SECOND	9
	{ .n_name = "_time_second" },
#define X_TIME		10
	{ .n_name = "_time" },
#define	X_NL_SIZE	11
	{ .n_name = NULL },
};
#endif

/*const char * _malloc_options = "XAJVP";*/
extern int paddr_resolv;
extern int af;
extern int debug_ioctl;
extern int hformat;

int verbose = 0;
int force = 0;
int debug = 0;
int clearconf = 0;
int help = 0;
int repeat = 0;
int traffic = 0;

static struct option longopts[] = {
	{ "add",              no_argument,            NULL,           'A' },
	{ "brother",          no_argument,            NULL,           'B' },
	{ "clone",            no_argument,            NULL,           'C' },
	{ "delete",           no_argument,            NULL,           'D' },
	{ "hashtable",        required_argument,      NULL,           'h' },
	{ "scheduler",        required_argument,      NULL,           's' },
	{ "timeout",          required_argument,      NULL,           't' },
	{ "timeout-establish", required_argument,     NULL,           1025 },
	{ "timeout-close",    required_argument,      NULL,           1026 },
	{ "timeout-syn",      required_argument,      NULL,           1027 },
	{ "active",           no_argument,            NULL,           'a' },
	{ "inactive",         no_argument,            NULL,           'i' },
	{ "persistent",       required_argument,      NULL,           'p' },
	{ "ssl",              required_argument,      NULL,           'p' },
	{ "tls",              required_argument,      NULL,           'p' },
	{ "rate",             required_argument,      NULL,           'r' },
	{ "weight",           required_argument,      NULL,           'w' },
	{ "clear",            no_argument,            &clearconf,         1 },
	{ "force",            no_argument,            &force,         1 },
	{ "help",             no_argument,            &help,          1 },
	{ "nohuman",          no_argument,            &hformat,       0 },
	{ "resolv",           no_argument,            NULL,           'n' },
	{ "verbose",          no_argument,            NULL,           'v' },
	{ "debug",            no_argument,            NULL,           'd' },
	/*{ "device",         required_argument,      NULL,           'd' },*/
	{ NULL,               0,                      NULL,           0 }
};




/***
 * get timeout
 */
char * get_timeout(char *string, int *data)
{
	char *str;

	if(!string) {
		return NULL;
	}

	str = strchr(string, ',');
	if(str) {
		*str++ = '\0';
		if(!is_digit(string)) {
			printf("Invalid timeout %s: Timeout must be a decimal integer.\n", string);
			exit(EXIT_FAILURE);
		}
		*data = atoi(string);
	} else {
		if(is_digit(string)) {
			*data = atoi(string);
			if(debug) printf("Input Read timeout: %d\n", *data);
			return NULL;
		}
		*data = 0;
		return NULL;
	}

	if(debug) printf("Input Read timeout: %d\n", *data);

	return str;
}





void quit(int s)
{
	endwin();
        puts("");
	exit(0);
}



void curses_init(void)
{
	initscr();
	start_color();
	curs_set(0);
	erase();
}



/* TODO a optimiser en bornant */
#define COMPUTE(NAME)										\
unsigned long compute_ ## NAME (int max, int id, struct lbdr_stats *ptr)			\
{												\
	unsigned long data = 0;									\
	int i;											\
												\
	for(i = 0; i < max; i++) {								\
		if(((ptr[i].type >> 16) == id) && ((ptr[i].type & 0xff) == LBDR_TYPE_SERVER)) {	\
			data += ptr[i].stats.srv.NAME;						\
		}										\
	}											\
	return data;										\
}

COMPUTE(ls_bytes);
COMPUTE(ls_pkts);
COMPUTE(ls_nb_cnx);
COMPUTE(ls_no_fin);
COMPUTE(ls_act_cnx);
COMPUTE(ls_inact_cnx);
COMPUTE(ls_shared_cnx);
#define PRINTIF(DATA, STR) if(DATA) printf(" %s", human(DATA, STR, HU_DIVISOR_1000))


/**
 * Print statistic ala ipvsadm or iostat
 */
int lbdr_print_stats(int n, struct lbdr_param *prm, struct sockaddr_storage *wanted_addr)
{
	unsigned int current_id = 0;
	int locked_id = -1;
	struct lbdr_stats *ptr;
	int i, tmp;

	ptr = prm->data;

	for(i = 0; i < n; i++) {
		current_id = ptr[i].type >> 16;

		if(wanted_addr) {
			if(locked_id == -1) {
				if(same_addr(wanted_addr, &(ptr[i].ip))) {
					if(((ptr[i].type) & 0xff) == LBDR_TYPE_SERVICE)
						locked_id = current_id;
				} else continue;
			} else {
				if(locked_id != current_id)
					continue;
			}
		}


		switch( (ptr[i].type) & 0xff ) {
			case LBDR_TYPE_SERVICE:
				printf("\n%s", paddr(&(ptr[i].ip)));
				if(verbose) {
					printf("%s", ptr[i].stats.svi.ls_persist ? " SSL" : "");
					printf(" %s scheduler,", sched_name(ptr[i].stats.svi.ls_sched, 0));
					printf(" Htable: %s,", human(ptr[i].stats.svi.ls_hashtablesize, "", HU_NOSPACE));
					printf(" Timeout: %d / %d / %d sec\n*** Total:",
						ptr[i].stats.svi.ls_timeout_syn,
						ptr[i].stats.svi.ls_timeout_est,
						ptr[i].stats.svi.ls_timeout_clo);
					printf("%c[%dG", 27, 30);
					printf("   %s", human(compute_ls_nb_cnx(n, current_id, ptr), " cnx", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(compute_ls_pkts(n, current_id, ptr), " pkts", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(compute_ls_bytes(n, current_id, ptr), " bytes", HU_DECIMAL|HU_NOSPACE));
					/* TODO: test % : */
					tmp = compute_ls_nb_cnx(n, current_id, ptr);
					if(tmp) 
						PRINTIF(ptr[i].stats.svi.ls_coll * 100 / tmp, " %Coll");
					PRINTIF(ptr[i].stats.svi.ls_err_nosyn, " NoSyn");
					PRINTIF(ptr[i].stats.svi.ls_err_nosrv, " Nosrv");
				} else {
					printf("%c[%dG", 27, 26);
					printf("   %s", human(compute_ls_nb_cnx(n, current_id, ptr), " cnx", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(compute_ls_act_cnx(n, current_id, ptr), " act", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(compute_ls_inact_cnx(n, current_id, ptr), " ina", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(compute_ls_shared_cnx(n, current_id, ptr), " shr", HU_DECIMAL|HU_NOSPACE));
				}
				puts("");
				break;
			case LBDR_TYPE_SERVER:
				printf("-%c %s", ptr[i].stats.srv.ls_status ? '>' : '|', paddr(&(ptr[i].ip)));
				printf(" [%d", ptr[i].stats.srv.ls_weight);
				if(verbose) {
					printf("/%s]", human(ptr[i].stats.srv.ls_rate, "", HU_NOSPACE|HU_DECIMAL|HU_DIVISOR_1000));
					PRINTIF(ptr[i].stats.srv.ls_err_unreach, " Unre");
					PRINTIF(ptr[i].stats.srv.ls_err_noroute, " Noro");
				} else printf("]");
				printf("%c[%dG", 27, verbose ? 30 : 26);
				printf("   %s", human(ptr[i].stats.srv.ls_nb_cnx, " cnx", HU_DECIMAL|HU_NOSPACE));

				if(traffic) {
					printf("   %s", human(ptr[i].stats.srv.ls_pkts, " pkts", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(ptr[i].stats.srv.ls_bytes, " bytes", HU_DECIMAL|HU_NOSPACE));
				} else {
					printf("   %s", human(ptr[i].stats.srv.ls_act_cnx, " act", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(ptr[i].stats.srv.ls_inact_cnx, " ina", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(ptr[i].stats.srv.ls_shared_cnx, " shr", HU_DECIMAL|HU_NOSPACE));
					printf("   %s", human(ptr[i].stats.srv.ls_no_fin, " nof", HU_DECIMAL));
				}

				puts("");
				break;
			case LBDR_TYPE_CLONE:
				printf("Clone: %s\n", paddr(&(ptr[i].ip)));
				break;
			default:
				printf("0x%x: %s\n", (ptr[i].type), paddr(&(ptr[i].ip)));
				break;
		}
	}

	return 0;
}

/** 
 * The same but manage diff between timelaps
 */
int lbdr_print_diff_stats(int n, struct lbdr_param *prm, struct sockaddr_storage *wanted_ip)
{
	return 0;
}
void print_load(void)
{
	double v[3] = { 0, 0, 0 };

	if(getloadavg(v, 3) == -1) {
		perror("getloadavg");
	}

	printw("Load avg: %.2f %.2f %.2f", v[0], v[1], v[2]);
}

void print_cpu_used(void)
{
	static long old_cp_time[CPUSTATES];
	long cp_time[CPUSTATES];
	long used = 0;
	size_t len = sizeof(cp_time);

	if(sysctlbyname("kern.cp_time", &cp_time, &len, NULL, 0) < 0) {
		fprintf(stderr, "Cannot get kern.cp_time");
	}

	used = cp_time[CP_USER] + cp_time[CP_NICE] + cp_time[CP_SYS] + cp_time[CP_INTR] + cp_time[CP_IDLE] 
	       - old_cp_time[CP_USER] - old_cp_time[CP_NICE] - old_cp_time[CP_SYS] - old_cp_time[CP_INTR] - old_cp_time[CP_IDLE];


	printw("CPU:%5.1f%% user,%5.1f%% nice,%5.1f%% system,%5.1f%% interrupt,%5.1f%% idle",
		(double)(cp_time[CP_USER] - old_cp_time[CP_USER]) * 100.0 / (double)used,
		(double)(cp_time[CP_NICE] - old_cp_time[CP_NICE]) * 100.0 / (double)used,
		(double)(cp_time[CP_SYS] - old_cp_time[CP_SYS])   * 100.0 / (double)used,
		(double)(cp_time[CP_INTR] - old_cp_time[CP_INTR]) * 100.0 / (double)used,
		(double)(cp_time[CP_IDLE] - old_cp_time[CP_IDLE]) * 100.0 / (double)used);

	memcpy(&old_cp_time, &cp_time, sizeof(cp_time));
}

#ifdef __FreeBSD__
void print_mem_used(int curses)
{
	struct memory_type_list *mtlp;
	struct memory_type *mtp;
	static kvm_t *kd;
	int error;

	mtlp = memstat_mtl_alloc();
	if (mtlp == NULL) {
		printf("memstat_mtl_alloc");
		return;
	}
	if (kd == NULL) {
		if (memstat_sysctl_malloc(mtlp, 0) < 0) {
			printf("memstat_sysctl_malloc: %s", memstat_strerror(memstat_mtl_geterror(mtlp)));
			return;
		}
	} else {
		if (memstat_kvm_malloc(mtlp, kd) < 0) {
			error = memstat_mtl_geterror(mtlp);
			if (error == MEMSTAT_ERROR_KVM)
				printf("memstat_kvm_malloc: %s", kvm_geterr(kd));
			else
				printf("memstat_kvm_malloc: %s", memstat_strerror(error));
		}
	}
	for (mtp = memstat_mtl_first(mtlp); mtp != NULL; mtp = memstat_mtl_next(mtp)) {
		
		/*
		if (memstat_get_numallocs(mtp) == 0 && memstat_get_count(mtp) == 0)
			continue;
			*/
		
		if(!memcmp("DRLB", memstat_get_name(mtp), 4)) {
			if(curses) {
				printw("drlb memory: %s used ",
					human(memstat_get_bytes(mtp), "", HU_NOSPACE));
				printw("in %s alloc  ",
					human(memstat_get_count(mtp), "", HU_NOSPACE));
				printw("(total: %s requests), ",
					human(memstat_get_numallocs(mtp), "", HU_NOSPACE));
			} else {
				printf( "Memory Statistics:\n"
					"\tCurrent usage is %s bytes", human(memstat_get_bytes(mtp), "", HU_NOSPACE));
				printf(" in %s requests\n", human(memstat_get_count(mtp), "", HU_NOSPACE));
				printf("\tCumulated usage is %s", human(memstat_get_memalloced(mtp), "", HU_NOSPACE));
				printf(" bytes in %s alloc()\n", human(memstat_get_numallocs(mtp), "", HU_NOSPACE));
				printf("\tCumulated usage is %s bytes", human(memstat_get_memfreed(mtp), "", HU_NOSPACE));
				printf(" in %s free()\n", human(memstat_get_numfrees(mtp), "", HU_NOSPACE));
				printf("\tTotal failure: %li\n", (long)memstat_get_failures(mtp));
			}
			break;
		}
	}
	memstat_mtl_free(mtlp);
}
#endif

#ifdef __NetBSD__
int deref_kptr(const void *kptr, void *ptr, size_t len, const char *msg)
{

	if (*msg == '_')
		msg++;
	if (kvm_read(kd, (u_long)kptr, (char *)ptr, len) != len)
		printf("kptr %lx: %s: %s\n", (u_long)kptr, msg, kvm_geterr(kd));

	return 0;
}
int kread(struct nlist *nl, int nlx, void *addr, size_t size)
{
	const char *sym;

	sym = nl[nlx].n_name;
	if (*sym == '_')
		++sym;
	if (nl[nlx].n_type == 0 || nl[nlx].n_value == 0)
		printf("symbol %s not defined\n", sym);
	deref_kptr((void *)nl[nlx].n_value, addr, size, sym);

	return 0;
}


void print_mem_used(int curses)
{
	struct malloc_type ks, *ksp;
	//struct kmembuckets buckets[MINBUCKET + 16];
	char *nlistf = NULL, *memf = NULL;
	char errbuf[1024];
	char memname[64];
	int c;

	gid_t egid = getegid();
	(void)setegid(getgid());
	(void)setegid(egid);


	kd = kvm_openfiles(nlistf, memf, NULL, O_RDONLY, errbuf);
	if (kd == NULL) {
		printf("kvm_openfiles: %s\n", errbuf);
		return;
	}
	if (nlistf == NULL && memf == NULL)
		(void)setgid(getgid());

	if ((c = kvm_nlist(kd, namelist)) != 0) {
		int doexit = 0;
		if (c == -1)
			printf("kvm_nlist: %s %s", "namelist\n", kvm_geterr(kd));
		for (c = 0; c < sizeof(namelist) / sizeof(namelist[0])-1; c++)
			if (namelist[c].n_type == 0 &&
			    c != X_TIME_SECOND &&
			    c != X_TIME) {
				if (doexit++ == 0)
					(void)fprintf(stderr, "vmstat: undefined symbols:");
				(void)fprintf(stderr, " %s",
				    namelist[c].n_name);
			}
		if (doexit) {
			(void)fputc('\n', stderr);
			exit(1);
		}
	}

	/*
	kread(namelist, X_KMEMBUCKETS, buckets, sizeof(buckets));
	if (buckets[0].kb_calls == 0) {
		if(!curses) 
			printf("Kmem statistics are not being gathered by the kernel.\n");
		return;
	}
	*/

	for (kread(namelist, X_KMEMSTAT, &ksp, sizeof(ksp));
	     ksp != NULL; ksp = ks.ks_next) {
		deref_kptr(ksp, &ks, sizeof(ks), "malloc type");
		deref_kptr(ks.ks_shortdesc, memname, sizeof(memname), "malloc type name");
		if(!strncasecmp(memname, "DRLB", 4)) {
			if(curses) {
				printw("drlb memory: %s used ",
					human(ks.ks_memuse, "", HU_NOSPACE));
				printw("in %s alloc  ",
					human(ks.ks_inuse, "", HU_NOSPACE));
				printw("(total: %s requests)",
					human(ks.ks_calls, "", HU_NOSPACE));
			} else {
				printf( "Memory Statistics:\n"
					"\tCurrent usage is %s bytes", human(ks.ks_memuse, "", HU_NOSPACE));
				printf(" in %s requests\n", human(ks.ks_inuse, "", HU_NOSPACE));
				printf("\tMaximum usage is %s bytes", human(ks.ks_maxused, "", HU_NOSPACE));
				printf(" and total %s request()\n", human(ks.ks_calls, "", HU_NOSPACE));
			}
			break;
		}
	}
}


#endif

/**
 * Print statistic ala 'top' with ncurses
 */
int lbdr_curses_stats(int n, struct lbdr_param *prm)
{
	static struct lbdr_param old_prm = { NULL, 0 };
	static unsigned int current_id = 0;
	struct lbdr_stats *ptr, *old_ptr;
	struct tm *t;
	time_t ti;
	unsigned long cnx;
	int i, line = 6;

	ti = time(NULL);
	t = localtime(&ti);

	erase();

	/* print load */
	move(0, 0);
	print_load();

	/* Print data */
	move(0, 70);
	printw("%2d:%2d:%2d", t->tm_hour, t->tm_min, t->tm_sec);

	move(2, 0);
	print_cpu_used();

	move(3, 0);
	print_mem_used(1);

	move(line - 1, 0);
	printw("            IP:               CNX/S         PKT/S       BYTES/S         %%COLL");

	if(old_prm.size != prm->size) {
		old_prm.size = prm->size;
		old_prm.data = realloc(old_prm.data, prm->size);
	}

	ptr = prm->data;
	old_ptr = old_prm.data;

	for(i = 0; i < n; i++) {
		current_id = ptr[i].type >> 16;

		switch( (ptr[i].type) & 0xff ) {
			case LBDR_TYPE_SERVICE:
				cnx = compute_ls_nb_cnx(n, current_id, ptr) - compute_ls_nb_cnx(n, current_id, old_ptr);
				move(line, 0);
				printw("%17s%c", paddr(&(ptr[i].ip)), ptr[i].stats.svi.ls_persist ? 's' : ' ');
				printw(" %s", sched_name(ptr[i].stats.svi.ls_sched, 0));
				move(line, 21);
				printw(" %13s", human(cnx, "", HU_DECIMAL|HU_NOSPACE));
				move(line, 35);
				printw(" %13s",
					human(compute_ls_pkts(n, current_id, ptr) - compute_ls_pkts(n, current_id, old_ptr), "", HU_DECIMAL|HU_NOSPACE));
				move(line, 49);
				printw(" %13s",
					human(compute_ls_bytes(n, current_id, ptr) - compute_ls_bytes(n, current_id, old_ptr), "", HU_DECIMAL|HU_NOSPACE));
				move(line, 73);
				printw("%4.2f", cnx ? (ptr[i].stats.svi.ls_coll - old_ptr[i].stats.svi.ls_coll) / cnx * 100 : 0.0);
				line++;
				break;
			case LBDR_TYPE_SERVER:
				/* TODO print if verbose */
			case LBDR_TYPE_CLONE:
			default:
				break;
		}
	}

	memcpy(old_prm.data, prm->data, prm->size);
	old_prm.size = prm->size;

	refresh();

	return 0;
}



int main(int ac, char *av[])
{
	struct sockaddr_storage *ip1, *ip2;
	char *name;
	char port[10];
	int r, hashtablesize = 15, tls = 0, status = 1, rate = 65000, weight = 10, sched = LBDR_SCHED_RR, repeat = 5, options = 0, t_est = 60, t_syn = 5, t_clo = 120;
	long cmd = 0;
	int c;

	r = strlen(av[0]) - 1;

	while(av[0][r] != '/' && r >= 0)
		r--;

	if(r != 0) name = &av[0][r+1];
	else name = av[0];


	while((c = getopt_long(ac, av, "ABCDacdh:ilnp:r:s:t:vw:46", longopts, NULL)) != -1) {
		switch(c) {
		case 'A': /* Add */
			if(cmd && cmd != LBDRADDCONF) {
				printf("Error: Can't add configuration and %s\n", ioctl_type(cmd));
				exit(EXIT_FAILURE);
			}
			cmd = LBDRADDCONF;
			break;
		case 'B': /* Share with brother */
			if(cmd && cmd != LBDRSETSHARE) {
				printf("Error: Can't add brother and %s\n", ioctl_type(cmd));
				exit(EXIT_FAILURE);
			}
			cmd = LBDRSETSHARE;
			break;
		case 'D': /* Del */
			if(cmd && cmd != LBDRDELCONF) {
				printf("Error: Can't delete configuration and %s\n", ioctl_type(cmd));
				exit(EXIT_FAILURE);
			}
			cmd = LBDRDELCONF;
			break;
		case 'C': /* clone */
			if(cmd && cmd != LBDRCLONE) {
				printf("Error: Can't clone configuration and %s\n", ioctl_type(cmd));
				exit(EXIT_FAILURE);
			}
			cmd = LBDRCLONE;
			break;
		case 'a': /* Change and active a server */
			if(cmd && cmd != LBDRCHGCONF) {
				printf("Error: Can't change configuration and %s\n", ioctl_type(cmd));
				exit(EXIT_FAILURE);
			}
			if(options & LBDR_OPTIONS_SERVER_STATUS) {
				puts("Error: Can't have -a and -i.");
				exit(1);
			}
			options |= LBDR_OPTIONS_SERVER_STATUS;
			status = 1;
			cmd = LBDRCHGCONF;
			break;
		case 'i': /* Change and inactive a sever */
			if(cmd && cmd != LBDRCHGCONF) {
				printf("Error: Can't change configuration and %s\n", ioctl_type(cmd));
				exit(EXIT_FAILURE);
			}
			if(options & LBDR_OPTIONS_SERVER_STATUS) {
				puts("Error: Can't have -a and -i.");
				exit(1);
			}
			options |= LBDR_OPTIONS_SERVER_STATUS;
			status = 0;
			cmd = LBDRCHGCONF;
			break;
		case 'p': /* persistence */
			if(!strncasecmp("on", optarg, 2))
				tls = 1;
			else if(!strncasecmp("off", optarg, 3))
				tls = 0;
			else if(!strncasecmp("enable", optarg, 6))
				tls = 1;
			else if(!strncasecmp("disable", optarg, 7))
				tls = 0;
			else printf("Invalid flag %s for persistence\n", optarg);
			options |= LBDR_OPTIONS_SERVICE_PERSIST;
			break;
		case 'r': /* set rate */
			options |= LBDR_OPTIONS_SERVER_RATE;
			if(!is_digit(optarg)) {
				printf("Invalid rate %s: must be a decimal integer.\n", optarg);
			} else {
				rate = atoi(optarg);
				if(rate > 65000) {
					puts("Maximum rate is 65000 cnx/seconds");
					rate = 65000;
				}
				if(rate < 1) {
					puts("Minimum rate is 1 cnx/seconds");
					rate = 1;
				}
			}
			break;
		case 'w': /* set weight */
			options |= LBDR_OPTIONS_SERVER_WEIGHT;
			if(!is_digit(optarg)) {
				printf("Invalid weight %s: must be a decimal integer.\n", optarg);
			} else {
				weight = atoi(optarg);
				if(weight > 100) {
					puts("Maximum weight is 100");
					weight = 100;
				}
				if(weight < 1) {
					puts("Minimum weight is 1");
					weight = 1;
				}
			}
			break;
		case 's': /* sched */
			options |= LBDR_OPTIONS_SERVICE_SCHED;
			sched = sched_id(optarg);
			if(sched == -1) {
				puts("Invalid Scheduler. Valid value are: rr, lc, wrr, wlc.");
				exit(EXIT_FAILURE);
			}
			break;

		case 't': /* timeout */
			/*
			if(cmd && cmd != LBDRCHGCONF) {
				printf("Error: Can't change configuration and %s\n", ioctl_type(cmd));
				exit(EXIT_FAILURE);
			}

			cmd = LBDRCHGCONF;
			*/
			optarg = get_timeout(optarg, &t_syn);
			optarg = get_timeout(optarg, &t_est);
			optarg = get_timeout(optarg, &t_clo);
			if(t_syn > 0) {
				if(debug) printf("Changing timeout SYN: %d\n", t_syn);
				options |= LBDR_OPTIONS_SERVICE_TIMEOUT|LBDR_OPTIONS_TIMEOUT_SYN;
			}
			if(t_est > 0) {
				if(debug) printf("Changing timeout EST: %d\n", t_est);
				options |= LBDR_OPTIONS_SERVICE_TIMEOUT|LBDR_OPTIONS_TIMEOUT_EST;
			}
			if(t_clo > 0) {
				if(debug) printf("Changing timeout CLO: %d\n", t_clo);
				options |= LBDR_OPTIONS_SERVICE_TIMEOUT|LBDR_OPTIONS_TIMEOUT_CLO;
			}
			break;
		case 1025:
			if(!is_digit(optarg)) {
				printf("Invalid timeout-establish %s: must be a decimal integer.\n", optarg);
			} else {
				t_est = atoi(optarg);
				options |= LBDR_OPTIONS_TIMEOUT_EST|LBDR_OPTIONS_SERVICE_TIMEOUT;
			}
			break;
		case 1026:
			if(!is_digit(optarg)) {
				printf("Invalid timeout-close %s: must be a decimal integer.\n", optarg);
			} else {
				t_clo = atoi(optarg);
				options |= LBDR_OPTIONS_TIMEOUT_CLO|LBDR_OPTIONS_SERVICE_TIMEOUT;
			}
			break;
		case 1027:
			if(!is_digit(optarg)) {
				printf("Invalid timeout-syn %s: must be a decimal integer.\n", optarg);
			} else {
				t_syn = atoi(optarg);
				options |= LBDR_OPTIONS_TIMEOUT_SYN|LBDR_OPTIONS_SERVICE_TIMEOUT;
			}
			break;
		case 'h':
			if(cmd && cmd != LBDRADDCONF) {
				puts("Hashsize can only be specified when adding configuration (-A).");
				exit(EXIT_FAILURE);
			}
			if(!is_digit(optarg)) {
				puts("Invalid weight: must be a decimal integer.");
			} else {
				hashtablesize = atoi(optarg);
				if(hashtablesize > 25) {
					hashtablesize = 25;
					puts("Maximum Hash table size: 2^25");
				} else if(hashtablesize < 10) {
					hashtablesize = 10;
					puts("Minimum Hash table size: 2^10");
				}
			}
			break;
		case '4':
			af = AF_INET;
			break;
		case '6':
			af = AF_INET6;
			break;
		case 'l':
			if(cmd && cmd != LBDRSTATS) {
				printf("Error: Can't print configuration and %s\n", ioctl_type(cmd));
				exit(EXIT_FAILURE);
			}
			cmd = LBDRSTATS;
			break;
		case 'v':
			verbose = 1;
			traffic = 1;
			break;
		case 'n':
			paddr_resolv = 0;
			break;
		case 'd':
			debug = 1;
			debug_ioctl++;
			break;
		case 0:
			if(help) goto usage;
			if(clearconf) {
				cmd = LBDRCLEAR;
			}
			break;
		default:
usage:
			printf("Usage: %s <options> <dst_ip,port> [<srv_ip>]\n"
				"Options can be:\n"
				"For -ADCBairw you need to specify <address,port> and <server_address>:\n"
				"\t--add or -A                  Add/Create server in service.\n"
				"\t--delete or -D               Delete server for service.\n"
				"\t--clone or -C                Clone service.\n"
				"\t--brother or -B              Share conn table with other LB.\n"
				"\n"
				"\t--active or -a               Activate server.\n"
				"\t--inactive or -i             Inactivate server.\n"
				"\t--rate or -r <rate>          Set rate for max cnx/sec.\n"
				"\t--weight or -w <weight>      Set weight.\n"
				"\n"
				"For -stp you need to specify only <address,port>\n"
				"\t--scheduler or -s <sched>    Set scheduler. [scheduler:rr|lc|wrr|wlc]\n"
				"\t--timeout or -t [tsyn,[test,[tclo]]]  Set timeout\n"
				"\t--timeout-syn [tsyn]         Set syn timeout.\n"
				"\t--timeout-establish [test]   Set establish timeout.\n"
				"\t--timeout-clo [tclo]         Set closed timeout.\n"
				"\t--persistent or --tls or --ssl or -p [on/off]\n"
				"\t\t\t\t     Set SSL/TLS persistance\n"
				"\t--hashsize or -h [size]      Set hash table size.\n"
				"\t--ipv4 or -4                 Force IPv4.\n"
				"\t--ipv6 or -6                 Force IPv6.\n"
				"\t--nohuman                    Don't print number in human format.\n"
				"\t--resolv or -n               Don't resolv hostname.\n"
				"\t--list or -l                 Print conf and cnx statistic.\n"
				"\t--verbose or -v              Be verbose (print traffic stats with -s).\n"
				"\t--clear                      Clear all service and server.\n"
				"\t--force                      Don't ask for clear.\n"
				"\t--debug                      Print some debug info.\n"
				"<weight> is an integer between 1 and 100\n"
				"<rate> is an integer between 1 and 65000\n"
				"<hash> is an integer between 1 and 25:\n"
				"\t-> will be used in power of 2 and only useful when creating service.\n"
				"\n" , name );
			return EXIT_FAILURE;
			break;
		}
	}

	if(options & LBDR_OPTIONS_SERVICE && cmd == 0)
		cmd = LBDRCHGCONF;


	ac -= optind;
	av += optind;

	if(!strncmp(name, "drlbstat", 8)) {
		struct lbdr_param *p;

		if(ac > 1) {
			puts("drlbstat doesn't take any parameters.");
			return -1;
		}

		debug = 0;
		verbose = 0;

		signal(SIGINT, quit);
		curses_init();

		for(;;) {
			r = lbdr_get_nip(0);
			p = lbdr_get_stats(r);
			lbdr_curses_stats(r, p);
			free(p);
			/*
			while(c != 'q') {
				c = getch();
				switch(c) {
					case 'q': endwin(); exit(0); break;
					default:
						printw("?");
				}
			}
			*/
			sleep(1);
		}
		return 0;
	}

	if(debug) {
		if(af == AF_INET) puts("Using IPv4 address (AF_INET)");
		else if(af == AF_INET6) puts("Using IPv6 address (AF_INET6)");
		else puts("Using unspecified address (PF_UNSPEC)");
	}

	switch(cmd) {
		case LBDRADDCONF:
			options |= LBDR_OPTIONS_TIMEOUT|LBDR_OPTIONS_SERVER;
		case LBDRCHGCONF:
		case LBDRCLONE:
		case LBDRDELCONF:
			if(ac < 1) {
				puts("Not enough parameter.");
				return -1;
			}
			ip1 = get_ip(av[0], "80", af);

			if(!ip1) {
				printf("Bad addr %s\n", av[0]);
				return -1;
			}

			if(  cmd == LBDRCHGCONF &&
			    (options & LBDR_OPTIONS_SERVICE) &&
			   !(options & LBDR_OPTIONS_SERVER)) {
				lbdr_change_server(cmd, ip1, NULL, hashtablesize, sched,
						status, weight, rate, tls, options,
						t_syn, t_est, t_clo);
				break;
			}

			if(ac < 2) {
				puts("Not enough parameter.");
				return -1;
			}

			snprintf(port, 10, "%d", ntohs(SPORT(ip1)));

			ip2 = get_ip(av[1], port, af);

			if(!ip2) {
				printf("Bad addr %s\n", av[1]);
				return -1;
			}

			lbdr_change_server(cmd, ip1, ip2, hashtablesize, sched, status, weight, rate, tls, options, t_syn, t_est, t_clo);

			break;

		case LBDRSETSHARE:
			if(ac != 2) {
				puts("Need IP and key for sharing.");
				return -1;
			}
			lbdr_set_brother(av[0], av[1]);
			break;
		case LBDRCLEAR:
			{
				int on = 1;
				if(!force) {
					printf("Are you sure do you want to clear all rules ? y/n : ");
					r = getchar();
					if(!(r == 'y' || r == 'Y')) return -1;
				}

				lbdr_ioctl(LBDRCLEAR, &on);
			}
			break;
		case LBDRSTATS:
			{
				struct lbdr_param *p;
				struct sockaddr_storage *ip = NULL;
				int on = 1;

				if(ac == 1) {
					ip = get_ip(av[0], "80", af);
					if(!ip) {
						printf("Bad parameter %s\n", av[0]);
					}
				}
				if(ac == 2 || (ac == 1 && ip == NULL)) {
					if(!is_digit(av[1])) {
						puts("interval must be a decimal integer.");
						return -1;
					}
					repeat = atoi(av[0]);
					if(verbose) printf("Repeat stats %i secondes\n", repeat);

					for(;;) {
						r = lbdr_get_nip(0);
						p = lbdr_get_stats(r);
						lbdr_print_diff_stats(r, p, ip);
						free(p);
						sleep(repeat);
					}
				} 
				if(debug)
					lbdr_ioctl(LBDRDEBUG, &on);

				r = lbdr_get_nip(verbose && (ip ? 0 : 1));
				p = lbdr_get_stats(r);
				if(verbose && !ip) print_mem_used(0);
				lbdr_print_stats(r, p, ip);
				free(p);
			}
			break;
		default:
			goto usage;
			break;
	}

	return 0;
}


/* EOF */
