/*-
 * Copyright (c) 2008 Jean-Yves Moulin <jym@ldh.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/* Temporary: 
#define DEBUGTIME 1
#define DEBUGTRACE 1
#define DEGROBUG  1
#define DDEBUG 1
*/



#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/malloc.h>
#include <sys/time.h>
#include <sys/systm.h>
#include <sys/ioccom.h>
#include <sys/conf.h>
/*#include <sys/mutex.h>*/
#include <sys/syslog.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/kthread.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/mbuf.h>
#include <net/if.h>
#include <net/pfil.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip6.h>
#include <netinet6/ip6_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/tcp.h>

#ifdef __NetBSD__
#include <sys/cdefs.h>
#include <machine/param.h>
/*#include <sys/lkm.h>*/
#include <machine/limits.h>
#if (__NetBSD_Version__ >= 399002000)
# include <sys/kauth.h>
#endif
#else /* if FreeBSD */
#include <sys/libkern.h>
#include <sys/limits.h>
#include <machine/in_cksum.h>
#endif


#ifdef DEBUGPRED
#include "pred.h"
#else
# define unlikely(expr) __builtin_expect (!!(expr), 0)
# define likely(expr) __builtin_expect (!!(expr), 1)
#endif

#include "lbdr.h"


/***********
 * MACRO
 */


/*******************************************/
/*************  DEBUG ONLY  ****************/
/*******************************************/

/* Sorry for these macros but I'm borring with #ifdef ... #endif */
#ifdef DDEBUG
   char				log_debug[2048];
   int				log_debug_ptr = 0;
#  define ADDLOG(...)		log_debug_ptr += snprintf(log_debug + log_debug_ptr, 2048 - log_debug_ptr, __VA_ARGS__)
#  define PRINT_LOG(X)		log(LOG_DEBUG, log_debug); log_debug_ptr = 0;
#  define RESET_LOG(X)		log_debug_ptr = 0;
#  define LIMIT_LOG 1
#  define D(X) X

#  ifdef DEGROBUG
#     define DG(X) X
#  endif

/* used for tracing all received packets */
#  ifdef DEBUGTRACE
     char			log_trace[2048];
     int			log_trace_ptr = 0;
#    define TRACE(...)		log_trace_ptr += snprintf(log_trace + log_trace_ptr, 2048 - log_trace_ptr, __VA_ARGS__)
#    define PRINT_TRACE(X)	log(LOG_DEBUG, log_trace); log_trace_ptr = 0;
#    define DTR(X) X
#  endif

#endif /* DDEBUG */



#ifdef DEBUGTIME
#  define INIT(X, Y)	X.ns = 0 ; X.nb = 0;			\
			F(if(!mtx_initialized(&(X.mtx))))  	\
			mtx_init(&(X.mtx), Y, NULL, MTX_DEF);

#  define START(X)	mtx_lock(&(X.mtx));			\
			X.tmp = read_tsc();

#  define STOP(X)	X.ns += (read_tsc() - X.tmp);		\
			X.nb++;					\
			mtx_unlock(&(X.mtx));

#  define DEINIT(X)	mtx_destroy(&(X.mtx));
#endif /* DEBUGTIME */




#ifndef D
# define D(X)
# define ADDLOG(...)
# define PRINT_LOG(X)
#endif
#ifndef DG
# define DG(X)
#endif
#ifndef DTR
# define DTR(X)
# define TRACE(X, ...)
# define PRINT_TRACE(X)
#endif
#ifndef INIT
# define INIT(X, Y)
#endif
#ifndef START
# define START(X)
#endif
#ifndef STOP
# define STOP(X)
#endif
#ifndef DEINIT
# define DEINIT(X)
#endif

/*******************************************/
/*************  END DEBUG  *****************/
/*******************************************/



#define FOREACHCLONE(X, ...) {					\
		struct lbdr_service *clone;			\
		struct lbdr_service *nextclone = X->clone;	\
		while(nextclone) {				\
			clone = nextclone;			\
			nextclone = clone->clone;		\
			__VA_ARGS__;				\
		} }

#define SPORT(ip)	ip->ss_family == AF_INET ? 		\
			((struct sockaddr_in *)ip)->sin_port :	\
			((struct sockaddr_in6 *)ip)->sin6_port

 
#ifdef NOBIG
#  define IF_BIG(X)		
#  define IF_NOBIG(X)		X
#  define MTX_LOCK()		mtx_lock(mtx)
#  define MTX_UNLOCK()		mtx_unlock(mtx)
#else /* ifdef BIG */
#  define IF_BIG(X)		X
#  define IF_NOBIG(X)	
#  define MTX_LOCK()		mtx_lock(&big_lock)
#  define MTX_UNLOCK()		mtx_unlock(&big_lock)
#endif

/*********************************************/
/*************  END MACRO  *******************/
/*********************************************/


#define LBDR_VERSION "1.0-alpha ("__DATE__" "__TIME__")"

/* hash table size **** BE CAREFULL: must be a power of 2 **** */

#ifndef LIMIT_LOG
#define LIMIT_LOG	500	/* Print syslog every X occurences */
#endif

#define LIMIT_SERVICE   128	/* power of Two < 4096 */
#define LIMIT_CNX     65000	/* Max limit to X cnx/second */
#define LIMIT_SERVER  32767	/* 2^15. Don't touch! */
#define OUT_UDP_PORT  10179	/* default port for sharing conn if not
				 * specified (why these one ? See lego) */

/* parametres par defaut: */
#define TIMEOUT_SYN	  5	/* secondes avant de passer en inactive */
#define TIMEOUT_ESTABL	 60	/* secondes avant de passer en inactive */
#define TIMEOUT_CLOSE	120	/* cnx inactive, sec avant de virer de la HT */

/* enum cnx_status { CNX_SYN, CNX_ESTABL, CNX_TIMEOUT, CNX_CLOSE, CNX_SYNTO, CNX_SHARED }; */
#define CNX_FLAG_OPEN   0x01
#define CNX_FLAG_DATA   0x02
#define CNX_FLAG_ACT    0x04
#define CNX_SYNTO	0x01 /* !CNX_FLAG_ACT  && !CNX_FLAG_DATA  &&  CNX_FLAG_OPEN */
#define CNX_CLOSE       0x02 /* !CNX_FLAG_ACT  &&  CNX_FLAG_DATA  && !CNX_FLAG_OPEN */
#define CNX_TIMEOUT     0x03 /* !CNX_FLAG_ACT  &&  CNX_FLAG_DATA  &&  CNX_FLAG_OPEN */
#define CNX_SYN         0x05 /*  CNX_FLAG_ACT  && !CNX_FLAG_DATA  &&  CNX_FLAG_OPEN */
#define CNX_ESTABL      0x07 /*  CNX_FLAG_ACT  &&  CNX_FLAG_DATA  &&  CNX_FLAG_OPEN */
#define CNX_SHARED	0x06 /*  CNX_FLAG_ACT  &&  CNX_FLAG_DATA  && !CNX_FLAG_OPEN */

/* enum srv_status { SRV_UP, SRV_DOWN }; */
#define SRV_DOWN        0x00
#define SRV_UP          0x01


/* enum mbuf_status { NOTHING, DROP }; */
#define NOTHING		0x01
#define DROP		0x02

/****
 * OS dependencies
 */

#ifdef __NetBSD__
/*******************************************/
/*************  NetBSD  ********************/
/*******************************************/

# define N(X) X
# define F(X)
# define __INT_MAX INT_MAX
# define CSUM_UDP M_CSUM_UDPv4
typedef void * caddr_t;
# define ip_newid()		ip_newid(NULL)
# define mtx_lock(X)		mutex_enter(X)
# define mtx_init(X, Y, Z, W)	mutex_init(X, MUTEX_DEFAULT, IPL_NONE)
# define mtx_unlock(X)		mutex_exit(X)
# define mtx_destroy(X)		mutex_destroy(X)
# define callout_deactivate(X)	callout_ack(X)
# define callout_sched(CAL, TICK, FUNC, ARG)  callout_schedule(CAL, TICK)
# define in_pseudo(X, Y, Z)	in_cksum_phdr(X, Y, Z)
# define rtalloc1(X, Y, Z)	rtalloc1(X, Y);
/* FIXME: not a very good solution ... : */
# define callout_drain(X)	if(callout_expired(X)) { \
					callout_stop(X); \
					callout_destroy(X); }

void drlbattach(int);
void drlbdetach(int);

/*int drlb_lkmentry(struct lkm_table *, int, int);*/

extern struct cfdriver drlb_cd;



#else /* __FreeBSD__ */
/*******************************************/
/*************  FreeBSD  *******************/
/*******************************************/

# define F(X) X
# define N(X)
# define CALLOUT_MPSAFE 1
typedef struct mtx kmutex_t;
static struct cdev *sdev;
int drlbioctl(struct cdev *dev, u_long cmd, caddr_t arg, int mode, struct thread *td);
# define callout_sched(CAL, TICK, FUNC, ARG) callout_reset(CAL, TICK, FUNC, ARG)

static struct cdevsw drlb_devsw = { 
	.d_version = D_VERSION,
	.d_ioctl = drlbioctl,
	.d_name = "drlb"
};


#endif


/*******************************************/
/*************  STRUCT Types  **************/
/*******************************************/


struct lbdr_server {
        struct rtentry *rt;
        struct sockaddr_storage *ip;
        uint32_t active_cnx;
        uint32_t inactive_cnx;
	uint32_t shared_cnx;
        uint32_t weight:10,
		 status:1,
        	 notused:1,
        	 data:20;
	u_long   nb_cnx;
        u_long   pkts;
        u_long   bytes;
        u_long   no_fin;
        uint32_t cur_sec;
        uint32_t cnx_sec; /* 16 bits high: max (<65535) , 16 bits low: current */
        uint32_t err_unreach;
        uint32_t err_noroute;
};
					/*32bits*/
struct lbdr_conn {			/* 164o*/
	struct sockaddr_storage ip_src;	/* 128 */
	struct callout cal;		/*  16 */
	struct lbdr_server *srv;	/*   4 */
	struct lbdr_service *parent;	/*   4 */
	struct lbdr_conn *next;		/*   4 */ /* sll puisque risque de collisions dans HT */
	IF_NOBIG(kmutex_t *pmtx;)	/*   4 */
	uint32_t hash:25,		/*   4 */
		 status:3;		/*   . */
	uint32_t last_seen;
};


struct counters {
	uint32_t t_syn;
	uint32_t t_est;
	uint32_t t_clo;
	uint32_t nb_coll;
	uint32_t err_nosrv;
	uint32_t err_nosyn;
};

struct lbdr_service {
        struct callout cal;
        struct sockaddr_storage *ip;
        struct lbdr_service *next;		/* sll puisque risque de collisions dans HT */
	uint32_t refcnt:6,
		 htsize:26;
	uint32_t nb_server:15,
		 used_server:15,
		 cloned:1,
		 persist:1;
	IF_NOBIG(kmutex_t *locktable;)
        struct lbdr_conn *htable;
	int (*sched)(struct lbdr_server **, int, int);
        struct lbdr_server **server;
	struct counters *to;
	struct lbdr_service *clone;
};

struct lbdr_sharing {
	struct sockaddr_storage src;
	struct sockaddr_storage dst;
	struct sockaddr_storage srv;
	uint32_t status;
};


/*******************************************/
/*****************  DATA   *****************/
/*******************************************/

#define IP_BUF 46 /* IPv6 == 39 + ':' + port (max 5char) */
static char ip_buf[IP_BUF];

static struct lbdr_service **service_available = NULL;
static unsigned int nticks = 0;
static unsigned long stats_nomem = 0;
static unsigned long stats_notsame = 0;
static unsigned long stats_dropact = 0;
static unsigned long stats_nombuf = 0;
static unsigned long stats_badmbuf = 0;
static unsigned long stats_udpbadcsum = 0;
static unsigned long stats_udpbadlen = 0;
static unsigned short sync_port = htons(OUT_UDP_PORT); /* stored in network order */
static struct lbdr_server brother;
static uint32_t sync_key = 0;
//static uint32_t log_warning = 0;
struct pfil_head *pfh_inet4 = NULL, *pfh_inet6 = NULL;

IF_BIG(static kmutex_t big_lock;)

#ifdef DEBUGTIME
struct timelost {
	uint64_t ns;
	uint64_t tmp;
	unsigned int nb;
	kmutex_t mtx;
};
static struct timelost fwd_mb, nw_cnx, up_cnx, fd_srv, fd_svi, inc_udp, rcv_cnx, fd_cnx, dr_cnx, cl_cnx, cr_cnx, schd, hsh, snd_cnx, input;

static void init_timelost(void)
{
	INIT(inc_udp, "timeudp");
	INIT(rcv_cnx, "timercv");
	INIT(snd_cnx, "timesndcnx");
	INIT(fwd_mb, "timefwd");
	INIT(nw_cnx, "timenew");
	INIT(up_cnx, "timeold");
	INIT(fd_svi, "timefdsvi");
	INIT(fd_srv, "timefdsrv");
	INIT(fd_cnx, "timefdcnx");
	INIT(cl_cnx, "timeclcnx");
	INIT(dr_cnx, "timedrcnx");
	INIT(cr_cnx, "timecrcnx");
	INIT(schd, "timeschd");
	INIT(hsh, "sfhash");
	INIT(input, "input");
}
static void clear_timelost(void)
{
	DEINIT(inc_udp);
	DEINIT(fwd_mb);
	DEINIT(nw_cnx);
	DEINIT(up_cnx);
	DEINIT(fd_srv);
	DEINIT(fd_svi);
	DEINIT(fd_cnx);
	DEINIT(cl_cnx);
	DEINIT(dr_cnx);
	DEINIT(cr_cnx);
	DEINIT(rcv_cnx);
	DEINIT(snd_cnx);
	DEINIT(schd);
	DEINIT(hsh);
	DEINIT(input);
}
#endif

#ifdef DEBUGPRED
extern long int __start_predict_data;
extern long int __stop_predict_data;
extern long int __start_predict_line;
extern const char *__start_predict_file;
#endif

static MALLOC_DEFINE(M_BUF, "DRLB", "DRLB Memory Buffers");


/*******************************************/
/*****************  CODE   *****************/
/*******************************************/

static __inline__ uint64_t read_tsc(void)
{
        register uint64_t t;
        /*__asm__ volatile ("XOR eax, eax;");*/
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
        __asm__ volatile ("xorl %eax, %eax");
	__asm__ volatile ("RDTSC" : "=A" (t));

        return t;
}



static __inline__ uint32_t sfhash4 (struct sockaddr_storage *ip)
{
	struct sockaddr_in *in4 = (struct sockaddr_in *)ip;
	register uint32_t hash, tmp;

	START(hsh);

	hash  = (in4->sin_addr.s_addr & 0x0000ffff) ;
	tmp   = (((in4->sin_addr.s_addr & 0xffff0000) >> 16) << 11) ^ hash;

	hash  = (hash << 16) ^ tmp;
	hash += hash >> 11;
	hash += in4->sin_port;
	hash ^= hash << 11;
	hash += hash >> 17;
	hash ^= hash << 3;
	hash += hash >> 5;
	hash ^= hash << 4;
	hash += hash >> 17;
	hash ^= hash << 25;
	hash += hash >> 6;

	STOP(hsh);

	return hash;
}
static __inline__ uint32_t sfhash6 (struct sockaddr_storage *ip)
{
	struct sockaddr_in6 *in6 = (struct sockaddr_in6 *)ip;
	register uint32_t hash, tmp;

	START(hsh);

	hash   = (in6->sin6_addr.s6_addr16[0]) ;
	tmp    = (in6->sin6_addr.s6_addr16[1] << 11) ^ hash;
	hash   = (hash << 16) ^ tmp;
	hash  += hash >> 11;
	hash  += in6->sin6_addr.s6_addr16[2];
	tmp    = (in6->sin6_addr.s6_addr16[3] << 11) ^ hash;
	hash   = (hash << 16) ^ tmp;
	hash  += hash >> 11;
	hash  += in6->sin6_addr.s6_addr16[4];
	tmp    = (in6->sin6_addr.s6_addr16[5] << 11) ^ hash;
	hash   = (hash << 16) ^ tmp;
	hash  += hash >> 11;
	hash  += in6->sin6_addr.s6_addr16[6];
	tmp    = (in6->sin6_addr.s6_addr16[7] << 11) ^ hash;

	hash  = (hash << 16) ^ tmp;
	hash += hash >> 11;
	hash += in6->sin6_port;
	hash ^= hash << 11;
	hash += hash >> 17;
	hash ^= hash << 3;
	hash += hash >> 5;
	hash ^= hash << 4;
	hash += hash >> 17;
	hash ^= hash << 25;
	hash += hash >> 6;

	STOP(hsh);

	return hash;
}




/**
 * print IP address and port, v4 or v6
 * DO NOT USE TWICE IN A PRINTF (because buffer is static)
 */
static char *paddr(struct sockaddr_storage *ip)
{
	struct sockaddr_in *in = (struct sockaddr_in *)ip;
	struct sockaddr_in6 *in6 = (struct sockaddr_in6 *)ip;

	bzero(ip_buf, IP_BUF);

	if(ip->ss_family == AF_INET) {
		snprintf(ip_buf, IP_BUF, "%u.%u.%u.%u",
			((unsigned char *)&in->sin_addr.s_addr)[0],
			((unsigned char *)&in->sin_addr.s_addr)[1],
			((unsigned char *)&in->sin_addr.s_addr)[2],
			((unsigned char *)&in->sin_addr.s_addr)[3]);

	} else if(ip->ss_family == AF_INET6) {
		snprintf(ip_buf, IP_BUF, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
			((unsigned char *)&in6->sin6_addr.s6_addr)[0],
			((unsigned char *)&in6->sin6_addr.s6_addr)[1],
			((unsigned char *)&in6->sin6_addr.s6_addr)[2],
			((unsigned char *)&in6->sin6_addr.s6_addr)[3],
			((unsigned char *)&in6->sin6_addr.s6_addr)[4],
			((unsigned char *)&in6->sin6_addr.s6_addr)[5],
			((unsigned char *)&in6->sin6_addr.s6_addr)[6],
			((unsigned char *)&in6->sin6_addr.s6_addr)[7],
			((unsigned char *)&in6->sin6_addr.s6_addr)[8],
			((unsigned char *)&in6->sin6_addr.s6_addr)[9],
			((unsigned char *)&in6->sin6_addr.s6_addr)[10],
			((unsigned char *)&in6->sin6_addr.s6_addr)[11],
			((unsigned char *)&in6->sin6_addr.s6_addr)[12],
			((unsigned char *)&in6->sin6_addr.s6_addr)[13],
			((unsigned char *)&in6->sin6_addr.s6_addr)[14],
			((unsigned char *)&in6->sin6_addr.s6_addr)[15]);
	} else snprintf(ip_buf, IP_BUF, "[Bad AF]");

	return ip_buf;
}


static const char *pstatus(unsigned int s)
{
	static char buf[16];
	switch(s) {
		case CNX_SYN:
			return "syn";
		case CNX_SYNTO:
			return "syn-timeout";
		case CNX_ESTABL:
			return "establish";
		case CNX_TIMEOUT:
			return "timeout";
		case CNX_CLOSE:
			return "close";
		case CNX_SHARED:
			return "shared";
		default:
			snprintf(buf, 16, "[Bad:%d]", s);
			return buf;
	}
	return NULL;
}


#if  defined (DEGROBUG) || defined (DEBUGTRACE)


static const char *picmptype(unsigned int t)
{
	static char buf[8];

	switch(t) {
		case ICMP_ECHOREPLY:
			return "echo-reply";
		case ICMP_UNREACH:
			return "dest-unreach";
		case ICMP_REDIRECT:
			return "redirect";
		case ICMP_SOURCEQUENCH:
			return "packet-loss";
		case ICMP_TIMXCEED:
			return "ttl-exceed";
		case ICMP_PARAMPROB:
			return "ip-header-bad";
		case ICMP_ECHO:
			return "echo-request";
		default:
			snprintf(buf, 8, "%d", t);
			return buf;
	}
	return NULL;
}


DTR(
static const char *pproto(unsigned int p)
{
        static char buf[8];
        switch (p) {
                case IPPROTO_IP:
                        return "IP";
                case IPPROTO_UDP:
                        return "UDP";
                case IPPROTO_TCP:
                        return "TCP";
                case IPPROTO_ICMP:
                        return "ICMP";
                default:
                        snprintf(buf, 8, "%dP", p);
			return buf;
	}
	return NULL;
}
)

#endif

/**
 * Is this better than memcmp() ?
 */
static __inline__ int same_addr(struct sockaddr_storage *a, struct sockaddr_storage *b)
{
	struct sockaddr_in *in4a, *in4b;
	struct sockaddr_in6 *in6a, *in6b;

	if(a->ss_family != b->ss_family) return 0;

        switch (a->ss_family) {
        case AF_INET:
		in4a = (struct sockaddr_in *)a;
		in4b = (struct sockaddr_in *)b;
                if( in4a->sin_addr.s_addr == in4b->sin_addr.s_addr &&
                           in4a->sin_port == in4b->sin_port)
			 return 1;
                break;
        case AF_INET6:
        	in6a = (struct sockaddr_in6 *)a;
        	in6b = (struct sockaddr_in6 *)b;

                if( in6a->sin6_addr.s6_addr32[0] == in6b->sin6_addr.s6_addr32[0] &&
                    in6a->sin6_addr.s6_addr32[1] == in6b->sin6_addr.s6_addr32[1] &&
                    in6a->sin6_addr.s6_addr32[2] == in6b->sin6_addr.s6_addr32[2] &&
                    in6a->sin6_addr.s6_addr32[3] == in6b->sin6_addr.s6_addr32[3] &&
			         in6a->sin6_port == in6b->sin6_port)
			return 1;
        	break;
        }

	return 0;
}



/**
 * I do not use memcpy() because I want all data to be fill (len, family..)
 */
static __inline__ void copy_addr(struct sockaddr_storage *dst, struct sockaddr_storage *src)
{
	struct sockaddr_in *dst4, *src4;
	struct sockaddr_in6 *dst6, *src6;

        switch (src->ss_family) {
        case AF_INET:
		dst4 = (struct sockaddr_in *)dst;
		src4 = (struct sockaddr_in *)src;

		dst4->sin_family = AF_INET;
		dst4->sin_len = sizeof(struct sockaddr_in);
                dst4->sin_addr.s_addr = src4->sin_addr.s_addr;
		dst4->sin_port = src4->sin_port;
        	return;

        case AF_INET6:
        	dst6 = (struct sockaddr_in6 *)dst;
        	src6 = (struct sockaddr_in6 *)src;

		dst6->sin6_family = AF_INET6;
		dst6->sin6_len = sizeof(struct sockaddr_in6);
                dst6->sin6_addr.s6_addr32[0] = src6->sin6_addr.s6_addr32[0];
                dst6->sin6_addr.s6_addr32[1] = src6->sin6_addr.s6_addr32[1];
                dst6->sin6_addr.s6_addr32[2] = src6->sin6_addr.s6_addr32[2];
                dst6->sin6_addr.s6_addr32[3] = src6->sin6_addr.s6_addr32[3];
		dst6->sin6_port = src6->sin6_port;
		return;
	default:
		log(LOG_ERR, "drlb: invalid address on copy!\n");
		dst->ss_family = 0;
		break;
        }

	return;
}


static int is_any(struct sockaddr_storage *ip)
{
	struct sockaddr_in *in4;
	struct sockaddr_in6 *in6;

	if(ip->ss_family == AF_INET) {
		in4 = (struct sockaddr_in *)ip;
		if(in4->sin_addr.s_addr == INADDR_ANY)
			return 1;
		else
			return 0;
	} else if(ip->ss_family == AF_INET6) {
		in6 = (struct sockaddr_in6 *)ip;
		if(IN6_IS_ADDR_UNSPECIFIED(&(in6->sin6_addr)))
			return 1;
		else
			return 0;
	} else {
		log(LOG_ERR, "drlb: invalid address on is_any()!\n");
	}

	return 0;
}






/****
 * Rate limitation
 * A very simple limitation:
 * We keep the number of cnx per second for the current second.
 * If second changed counter is back to zero.
 * Else counter is increased
 * And cnx refused if counter bigger than limit.
 */

static __inline__ int rate_limit(uint32_t *s, uint32_t *c)
{
	uint32_t cur_conn = (*c) & 0x0000FFFF;
	uint32_t max_conn = (*c) >> 16;

	if(*s == time_second) {
		if(cur_conn >= max_conn) {
			TRACE("*");
			return 1;
		}
		cur_conn++;
		*c = (max_conn << 16 | cur_conn);
		return 0;
	}

	*s = time_second;
	cur_conn = 1;
	*c = (max_conn << 16 | cur_conn);

	return 0;
}

/****
 * Scheduler:
 *
 * RR
 * LC
 * WRR
 * WLC
 */

/* Round robin scheduleur */
static int scheduler_rr(struct lbdr_server **s, int nb, int last)
{
	register uint32_t i;

	START(schd);

	i = last;

	TRACE("[rr ");

	do {
		i++;
		i %= nb;
		if(s[i]->status == SRV_UP && !rate_limit(&(s[i]->cur_sec), &(s[i]->cnx_sec))) {
			TRACE("%i] ", i);
			STOP(schd);
			return i;
		}
	} while(i != last);

	TRACE("-1] ");
	STOP(schd);
	return -1;
}

/* least connections scheduleur */
static int scheduler_lc(struct lbdr_server **s, int nb, int last)
{
	register int choice = -1, i;

	START(schd);

	i = last;

	TRACE("[lc ");

	do {
		i++;
		i %= nb;
		if(s[i]->status == SRV_UP) {
			if(rate_limit(&(s[i]->cur_sec), &(s[i]->cnx_sec)))
				continue;
			if(choice == -1) {
				choice = i;
			} else if(s[i]->active_cnx < s[choice]->active_cnx) {
				choice = i;
			}
		}
	} while(i != last);

	TRACE("%i] ", choice);
	STOP(schd);
	return choice;
}


/* meme chose que RR mais prend en compte le poid
 * Round: nb de tour fait
 * Total: nb de tour a faire
 * weight: nb de fois qu'il devra etre appele apres total
 * Data: nb de fois qu'il a ete appele depuis round
 */

static int scheduler_wrr(struct lbdr_server **s, int nb, int last)
{
	register int i;
	register int choice = -1;
	register uint32_t round = 1;
	register uint32_t total = 0;

	START(schd);
	TRACE("[wrr ");

	for(i = 0; i < nb; i++) {
		total += s[i]->weight;
		round += s[i]->data;
	}

	if(round > total) {
		round = 1;
	}

	for(i = 0; i < nb; i++) {
		if(round == 1) s[i]->data = 0;
		if(s[i]->status == SRV_UP && s[i]->weight > 0) {
			if(rate_limit(&(s[i]->cur_sec), &(s[i]->cnx_sec)))
				continue;
			if(choice == -1) {
				if((s[i]->data * total) < (s[i]->weight * round)) {
					choice = i;
					s[i]->data++;
					break;
				}
			}
		}
	}

	TRACE("%i]", choice);
	STOP(schd);

	return choice;
}

/* Weighted Least Connections */
static int scheduler_wlc(struct lbdr_server **s, int nb, int last)
{
	register int tmp_load = 0;
	register int choice_load = __INT_MAX;
	register int choice = -1, i;

	i = last;

	START(schd);

	TRACE("[wlc ");

	do {
		i++;
		i %= nb;
		if(s[i]->status == SRV_UP && s[i]->weight > 0) {
			if(rate_limit(&(s[i]->cur_sec), &(s[i]->cnx_sec)))
				continue;
			tmp_load = (s[i]->active_cnx / s[i]->weight);
			if(choice == -1) {
				choice = i;
				choice_load = tmp_load;
			} else if(tmp_load < choice_load) {
				choice = i;
				choice_load = tmp_load;
			}
		}
	} while(i != last);

	TRACE("%i) ", choice);

	STOP(schd);
	return choice;
}



static int sched_number(int (*sched)(struct lbdr_server **, int, int))
{
	if(sched == scheduler_rr) {
		return LBDR_SCHED_RR;
	} else if(sched == scheduler_wrr) {
		return LBDR_SCHED_WRR;
	} else if(sched == scheduler_lc) {
		return LBDR_SCHED_LC;
	} else if(sched == scheduler_wlc) {
		return LBDR_SCHED_WLC;
	}

	return -1;
}



/***
 * manipulating service:
 *
 * register_service();
 * unregister_service();
 * find_service();
 * free_service();
 * clone_service();
 */


static struct lbdr_service *register_service(struct sockaddr_storage *ip, int sched, int persist, uint32_t htsize)
{
	struct lbdr_service *last = NULL;
	struct lbdr_service *tmp;
	int i;
	uint32_t h;

	DG(uprintf("drlb: Service %s.%i... ", paddr(ip), ntohs(SPORT(ip)));)

	if(htsize > 25) htsize = 25;
	else if(htsize < 1) htsize = 1;

	htsize = 1 << htsize;

	if(ip->ss_family == AF_INET6)
		h = sfhash6(ip) & (LIMIT_SERVICE-1);
	else if(ip->ss_family == AF_INET)
		h = sfhash4(ip) & (LIMIT_SERVICE-1);
	else {
		D(uprintf("drlb: Invalid service family (not IPv4 nor IPv6)\n");)
		return NULL;
	}

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, ip)) {
			DG(uprintf("Already alive.\n"));
			return tmp;
		}
		last = tmp;
		tmp = tmp->next;
	};

	DG(uprintf("No existent => ");)

	D(uprintf("New service %s(HConn: %i) ", persist ? "[SSL] " : "", htsize);)

	tmp = malloc(sizeof(struct lbdr_service), M_BUF, M_NOWAIT | M_ZERO);
	if(!tmp) goto out;

	tmp->ip = malloc(sizeof(struct sockaddr_storage), M_BUF, M_NOWAIT | M_ZERO);
	if(!tmp->ip) goto free_tmp;

	tmp->htable = malloc(sizeof(struct lbdr_conn) * htsize, M_BUF, M_NOWAIT | M_ZERO);
	if(!tmp->htable) goto free_ip;

	tmp->to = malloc(sizeof(struct counters), M_BUF, M_NOWAIT|M_ZERO);
	if(!tmp->to) goto free_htable;

	/*
	IF_NOBIG(tmp->locktable = malloc(sizeof(kmutex_t) * htsize, M_BUF, M_NOWAIT | M_ZERO);)
	IF_NOBIG(if(!tmp->locktable) goto free_to;)
	*/

	for(i = 0; i < htsize; i++) {
		IF_NOBIG(mtx_init(&(tmp->locktable[i]), "locktable", NULL, MTX_DEF);)
		tmp->htable[i].status = 0;
		callout_init(&(tmp->htable[i].cal), CALLOUT_MPSAFE);
	}

	copy_addr(tmp->ip, ip);
	tmp->htsize = htsize;
	tmp->server = NULL;
	tmp->next = NULL;
	tmp->clone = NULL;
	tmp->cloned = 0;
	tmp->persist = persist;
	tmp->nb_server = 0;
	tmp->to->nb_coll = 0;
	tmp->refcnt = 0;
	tmp->to->err_nosyn = 0;
	tmp->to->err_nosrv = 0;
	tmp->used_server = 0;
	tmp->to->t_syn = TIMEOUT_SYN * nticks;
	tmp->to->t_est = TIMEOUT_ESTABL * nticks;
	tmp->to->t_clo = TIMEOUT_CLOSE * nticks;
	callout_init(&(tmp->cal), CALLOUT_MPSAFE);

	switch(sched) {
		case LBDR_SCHED_WLC:
			D(uprintf("Using Weighted-Least-Connections scheduler\n");)
			tmp->sched = scheduler_wlc;
			break;
		case LBDR_SCHED_WRR:
			D(uprintf("Using Weighted-Round-Robin scheduler\n");)
			tmp->sched = scheduler_wrr;
			break;
		case LBDR_SCHED_LC:
			D(uprintf("Using Least-Connections scheduler\n");)
			tmp->sched = scheduler_lc;
			break;
		default:
			D(uprintf("Unknow scheduler, ");)
		case LBDR_SCHED_RR:
			D(uprintf("Using Round-Robin scheduler\n");)
			tmp->sched = scheduler_rr;
			break;
	}

	if(!service_available[h]) {
		service_available[h] = tmp;
	} else /*if(last)*/ {
		last->next = tmp;
	}

	return tmp;

IF_NOBIG(free_to:
	free(tmp->to, M_BUF);)
free_htable:
	free(tmp->htable, M_BUF);
free_ip:
	free(tmp->ip, M_BUF);
free_tmp:
	free(tmp, M_BUF);
out:
	log(LOG_ERR, "drlb: not enough kernel memory!\n");
	stats_nomem++;
	return NULL;
}



/***
 * Fonction appele pour chaque paquet TCP.
 * I hope this is fast enough...
 */

static struct lbdr_service *find_service4(struct sockaddr_storage *ip)
{
	struct lbdr_service *tmp;
	uint32_t h;

	START(fd_svi);

	h = sfhash4(ip) & (LIMIT_SERVICE-1);

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, ip)) {
			STOP(fd_svi);
			return tmp;
		}
		tmp = tmp->next;
	};

	STOP(fd_svi);
	return NULL;
}

static struct lbdr_service *find_service6(struct sockaddr_storage *ip)
{
	struct lbdr_service *tmp;
	uint32_t h;

	START(fd_svi);

	h = sfhash6(ip) & (LIMIT_SERVICE-1);

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, ip)) {
			STOP(fd_svi);
			return tmp;
		}
		tmp = tmp->next;
	};

	STOP(fd_svi);
	return NULL;
}

static __inline__ struct lbdr_service *find_service(struct sockaddr_storage *ip)
{
	if(ip->ss_family == AF_INET6)
		return find_service6(ip);

	return find_service4(ip);
}



/*
 * Must lock the structure ?
 */
static void free_service(void *arg)
{
	struct lbdr_service *tmp = (struct lbdr_service *)arg;
	struct lbdr_conn *tcnx;
	struct lbdr_conn *ncnx = NULL;
	int i;

#ifdef __NetBSD__
	for(i = 0; i < tmp->htsize; i++) {
		tcnx = &(tmp->htable[i]);
		ncnx = tcnx->next;
		if(tcnx->status) callout_schedule(&(tcnx->cal), 1);
		while(ncnx) {
			tcnx = ncnx;
			ncnx = tcnx->next;
			callout_schedule(&(tcnx->cal), 1);
		}
	}
#endif

	while(tmp->refcnt > 0) {
		/* TODO: test this */
		D(uprintf("drlb: service in use, re-schedule free()");)
		callout_reset(&(tmp->cal), 1, free_service, arg);
		return;
	}

	if(tmp->cloned) {
		free(tmp->ip, M_BUF);
		free(tmp, M_BUF);
		return;
	}


	/* delete clone */
	FOREACHCLONE(tmp,
		while(clone->refcnt > 0);
		free(clone->ip, M_BUF);
		free(clone, M_BUF);
	)

	for(i = 0; i < tmp->htsize; i++) {
		tcnx = &(tmp->htable[i]);
		ncnx = tcnx->next;
		/* TODO: test */
		if(tcnx->status) {
			callout_drain(&(tcnx->cal));
			uprintf("callout to free()");
		}
		while(ncnx) {
			tcnx = ncnx;
			ncnx = tcnx->next;
			D(if(tcnx->status == CNX_SYN) uprintf(".");
			  else if(tcnx->status == CNX_ESTABL)  uprintf("*");
			  else if(tcnx->status == CNX_CLOSE)  uprintf("+");
			  else if(tcnx->status == CNX_TIMEOUT)  uprintf("-");
			  else if(tcnx->status == CNX_SYNTO) uprintf("_");
			  else if(tcnx->status == CNX_SHARED) uprintf("=");
			  else uprintf("?");)
			callout_drain(&(tcnx->cal));
			free(tcnx, M_BUF);
		}
		IF_NOBIG(mtx_destroy(&(tmp->locktable[i]));)
	}
	for(i = 0; i < tmp->nb_server; i++) {
		if(tmp->server[i]) {
			D(uprintf("s"));
			if(tmp->server[i]->rt)
				RTFREE(tmp->server[i]->rt);
			free(tmp->server[i]->ip, M_BUF);
			free(tmp->server[i], M_BUF);
		}
	}
	IF_NOBIG(free(tmp->locktable, M_BUF);)
	free(tmp->htable, M_BUF);
	free(tmp->server, M_BUF);
	free(tmp->to, M_BUF);
	free(tmp->ip, M_BUF);
	free(tmp, M_BUF);
}





static int unregister_service(struct sockaddr_storage *ip)
{
	struct lbdr_service *last = NULL;
	struct lbdr_service *tmp;
	uint32_t h;

	DG(uprintf("drlb: Service %s.%i... ", paddr(ip), ntohs(SPORT(ip)));)

	if(ip->ss_family == AF_INET6)
		h = sfhash6(ip) & (LIMIT_SERVICE-1);
	else
		h = sfhash4(ip) & (LIMIT_SERVICE-1);

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, ip)) {
			break;
		}
		last = tmp;
		tmp = tmp->next;
	};

	if(!tmp) {
		DG(uprintf("Not found ?\n"));
		return -1;
	}

	D(uprintf("[S");)

	if(!tmp->cloned) {
		FOREACHCLONE(tmp,
			unregister_service(clone->ip);
		)
	}
	D(uprintf("]");)

	if(service_available[h] == tmp) {
		service_available[h] = tmp->next;
	} else /*if(last)*/ {
		last->next = tmp->next;
	}

	DG(uprintf(" Removed.\n"));

	return 0;
}



/*
 *
 */

static int clone_service(struct lbdr_service *service, struct sockaddr_storage *nip)
{
	struct lbdr_service *last = NULL;
	struct lbdr_service *tmp;
	struct lbdr_service *new;
	uint32_t h;

	DG(uprintf("drlb: Service %s.%i... ", paddr(nip), ntohs(SPORT(nip)));)

	if(nip->ss_family == AF_INET6)
		h = sfhash6(nip) & (LIMIT_SERVICE-1);
	else
		h = sfhash4(nip) & (LIMIT_SERVICE-1);

	tmp = service_available[h];

	while(tmp) {
		if(same_addr(tmp->ip, nip)) {
			DG(uprintf("Already alive.\n"));
			return EEXIST;
		}
		last = tmp;
		tmp = tmp->next;
	};

	DG(uprintf("No existent => ");)
	D(uprintf("New service clone of %s.%i\n", paddr(service->ip), ntohs(SPORT(service->ip)));)

	new = malloc(sizeof(struct lbdr_service), M_BUF, M_NOWAIT | M_ZERO);
	if(!new) goto exit;

	memcpy(new, service, sizeof(struct lbdr_service));

	new->ip = malloc(sizeof(struct sockaddr_storage), M_BUF, M_NOWAIT | M_ZERO);
	if(!new->ip) goto exit_new;

	copy_addr(new->ip, nip);
	new->cloned = 1;
	new->clone = NULL;
	new->next = NULL;

	if(!service_available[h]) {
		service_available[h] = new;
	} else /*if(last)*/ {
		last->next = new;
	}

	/* add to clone list */
	tmp = service->clone;

	if(!tmp) {
		service->clone = new;
	} else {
		while(tmp->clone) {
			tmp = tmp->clone;
		}
		tmp->clone = new;
	}

	return 0;

exit_new:
	free(new, M_BUF);
exit:
	return ENOMEM;
}



/***
 * manipulating servers:
 *
 * add_server();
 * delete_server();
 * modify_server();
 * find_server():
 */


/* tableau de pointeurs sur liste serveurs, nb de serveur, nelle ip */
static struct lbdr_server ** add_server(struct lbdr_server **server, int nb, struct sockaddr_storage *s, int weight, int rate, int *err)
{
	struct lbdr_server **srv;
	int i = 0;

	if((s->ss_family != AF_INET) && (s->ss_family != AF_INET6)) {
		D(uprintf("drlb: Invalid servers family (not IPv4 nor IPv6)\n");)
		*err = EFAULT;
		return NULL;
	}

	if(nb >= LIMIT_SERVER) {
		D(uprintf("drlb: Can't add more than %i servers\n", LIMIT_SERVER);)
		*err = EOVERFLOW;
		return NULL;
	}

	if(rate <= 0 || rate > LIMIT_CNX) {
		D(uprintf("drlb: cnx rate %d is invalid\n", rate);)
		*err = EOVERFLOW;
		return NULL;
	}

	DG(uprintf("drlb: Server %s.%i... ", paddr(s), ntohs(SPORT(s)));)

	while(i < nb) {
		if(server[i]) {
			if(same_addr(server[i]->ip, s)) {
				DG(uprintf("Already here.\n"));
				*err = EEXIST;
				return NULL;
			}
		}
		i++;
	}

	DG(uprintf("No existent => " );)
	D(uprintf("New server %s.%i ", paddr(s), ntohs(SPORT(s)));)

	srv = malloc(sizeof(struct lbdr_server *) * nb+1, M_BUF, M_NOWAIT | M_ZERO);
	if(!srv) goto out;

	for(i = 0; i < nb; i++) {
		srv[i] = server[i];
	}

	srv[nb] = malloc(sizeof(struct lbdr_server), M_BUF, M_NOWAIT | M_ZERO);
	if(!srv[nb]) goto free_srv;

	srv[nb]->ip = malloc(sizeof(struct sockaddr_storage), M_BUF, M_NOWAIT | M_ZERO);
	if(!srv[nb]->ip) goto free_srv_tab;

        srv[nb]->rt = NULL;
        srv[nb]->status = SRV_UP;
        srv[nb]->weight = weight;
        srv[nb]->inactive_cnx = 0;
        srv[nb]->active_cnx = 0;
        srv[nb]->nb_cnx = 0;
        srv[nb]->no_fin = 0;
        srv[nb]->pkts = 0;
        srv[nb]->bytes = 0;
        srv[nb]->err_unreach = 0;
        srv[nb]->cur_sec = time_second;
        srv[nb]->cnx_sec = (rate << 16);

	copy_addr(srv[nb]->ip, s);

	srv[nb]->rt = rtalloc1((struct sockaddr *)srv[nb]->ip, 1, 0UL);

	/* FIXME: really need to decrement ? */
	if(srv[nb]->rt) {
		/*N(srv[nb]->rt->rt_refcnt--;)*/
		F(RT_UNLOCK(srv[nb]->rt);)
	}

        if(!(srv[nb]->rt)) {
        	log(LOG_ERR, "drlb: Error: Host %s is unreachable!\n", paddr(s));
        } else if(srv[nb]->rt->rt_flags & RTF_GATEWAY) {
        	log(LOG_ERR, "drlb: Error: Host %s is not on my network!\n", paddr(s));
        }
        D(else uprintf("(route via %s, w:%i, limit:%i)\n", srv[nb]->rt->rt_ifp->if_xname, weight, rate));

	*err = 0;
	return srv;

free_srv_tab:
	free(srv[nb], M_BUF);

free_srv:
	free(srv, M_BUF);

out:
	log(LOG_ERR, "drlb: not enough kernel memory!\n");
	stats_nomem++;
	*err = ENOMEM;
	return NULL;
}





static int modify_server(struct lbdr_server **server, struct sockaddr_storage *s, int nb, int options, int weight, int rate, int status)
{
	int i = nb - 1;

	D(uprintf("drlb: Server %s.%i... ", paddr(s), ntohs(SPORT(s)));)

	while(i >= 0) {
		if(same_addr(server[i]->ip, s)) {

			if(options & LBDR_OPTIONS_SERVER_STATUS) { 
				server[i]->status = status;
				DG(uprintf(" change status");)
			}

			if(options & LBDR_OPTIONS_SERVER_WEIGHT) {
				if(weight < 0 || weight > 100) {
					D(uprintf("Invalid weight (%i). Weight must be between 0 and 100\n", weight);)
					return EINVAL;
				}
				DG(uprintf(" change weight");)
				server[i]->weight = weight;
			}

			if(options & LBDR_OPTIONS_SERVER_RATE) {
				if(rate <= 0 || rate > LIMIT_CNX)
					return EINVAL;
				server[i]->cnx_sec = (rate << 16) | (server[i]->cnx_sec & 0x0000ffff);
				DG(uprintf(" change rate");)
			}

			D(uprintf("\ndrlb: %s (w: %i,  %i cnx/sec).\n",
				server[i]->status ? "Active" : "Inactive",
				server[i]->weight, 
				server[i]->cnx_sec >> 16);)
			D(log(LOG_DEBUG, "drlb: changed: %s %s (w: %i,  %i cnx/sec).\n",
				server[i]->status ? "Active" : "Inactive",
				paddr(s),
				server[i]->weight,
				server[i]->cnx_sec >> 16);)
			return 0;
		}
		i--;
	}

	D(uprintf("not found.\n");)

	return ENOENT;
}





static struct lbdr_server ** delete_server(struct lbdr_server **server, struct sockaddr_storage *s, int nb, int *err)
{
	struct lbdr_server **srv;
	int i = 0, dst = 0;

	nb--;

	D(uprintf("Rebuilding server table [%i] without %s.%i... ", nb, paddr(s), ntohs(SPORT(s)));)

	srv = malloc(sizeof(struct lbdr_server *) * nb, M_BUF, M_NOWAIT | M_ZERO);
	if(!srv) {
		log(LOG_ERR, "drlb: not enough kernel memory!\n");
		stats_nomem++;
		*err = ENOMEM;
		return NULL;
	}

	while(i <= nb) {
		if(same_addr(server[i]->ip, s)) {
			if(server[i]->active_cnx || server[i]->inactive_cnx) {
				D(uprintf("Can't delete a server with actives connections. Please deactivate it before.\n");)
				free(srv, M_BUF);
				*err = EBUSY;
				return NULL;
			}
			if(server[i]->rt)
				RTFREE(server[i]->rt);
			free(server[i]->ip, M_BUF);
			free(server[i], M_BUF);
			server[i] = NULL;
		} else {
			if(dst < nb)
				srv[dst] = server[i];
			dst++;
		}
		i++;
	}

	D(uprintf("%s\n", dst != i ? "OK" : "FAIL"));

	if(dst != i) {
		*err = 0;
		return srv;
	}

	free(srv, M_BUF);
	*err = ENOENT;
	return NULL;
}


static int find_server(struct lbdr_server **server, int nb, struct sockaddr_storage *s)
{
	int i = nb - 1;

	START(fd_srv);

	while(i >= 0) {
		if(same_addr(server[i]->ip, s)) {
			STOP(fd_srv);
			return i;
		}
		i--;
	}

	STOP(fd_srv);
	return -1;
}

/***
 * manipulating connections:
 * create_cnx();
 * find_cnx();
 * close_cnx();
 * drop_cnx();
 */

/*static struct lbdr_conn * create_cnx(struct lbdr_conn *ht, uint32_t h, struct lbdr_server *server, struct sockaddr_storage *ip, u_int *coll, kmutex_t *p)*/

static struct lbdr_conn * create_cnx(struct lbdr_service *service, uint32_t h, int srv, struct sockaddr_storage *ip)
{
	struct lbdr_conn *cnx;

	START(cr_cnx);

	cnx = &(service->htable[h]);

	if(cnx->status != 0) {
		while(cnx->next) cnx = cnx->next;
		cnx->next = malloc(sizeof(struct lbdr_conn), M_BUF, M_NOWAIT);
		cnx = cnx->next;
		if(!cnx) {
			STOP(cr_cnx);
			return NULL;
		}
		cnx->next = NULL;
		service->to->nb_coll++;
	}

	callout_init(&(cnx->cal), CALLOUT_MPSAFE);
	cnx->status = CNX_SYN;
	cnx->srv = service->server[srv];
	cnx->parent = service;
	IF_NOBIG(cnx->pmtx = &(service->locktable[h]);)
	copy_addr(&(cnx->ip_src), ip);
	cnx->hash = h;
	cnx->last_seen = time_second;

	STOP(cr_cnx);
	return cnx;
}


static struct lbdr_conn *find_cnx(struct lbdr_conn *ht, uint32_t h, struct sockaddr_storage *ip)
{
	struct lbdr_conn *tmp;

	START(fd_cnx);

	tmp = &(ht[h]);

	while(tmp) {
		if(same_addr(&(tmp->ip_src), ip)) {
			STOP(fd_cnx);
			return tmp;
		}
		tmp = tmp->next;
	}
	
	STOP(fd_cnx);
	return NULL;
}


/**
 * output function
 * Do I need to check:
 *	- mbuf size (> struct ip[6]) ?
 *	- header lenght ?
 *	- checksum ? no!
 *	- packet lenght
 *	- traffic conditionner or firewall
 *	- fragment
 * I doesn't verify that pfil will do that for me..
 */
static __inline__ int fwd(struct lbdr_server *srv, struct mbuf *mb)
{
	struct rtentry *rt = srv->rt;

	START(fwd_mb);

	if (likely(rt)) {
		if(likely(rt->rt_ifp && (rt->rt_flags & RTF_UP)))
			goto fwd;
		RTFREE(srv->rt);
	}
	srv->rt = rtalloc1((struct sockaddr *)srv->ip, 1, 0UL);
	printf("RTALLOC for %s  Addr: %p\n", paddr(srv->ip), srv->rt);
	rt = srv->rt;

	if(unlikely(!rt)) {
		srv->err_noroute++;

		if(!((srv->err_noroute - 1) % LIMIT_LOG))
			log(LOG_ERR, "drlb: No route to host %s !\n", paddr(srv->ip));

		STOP(fwd_mb);
		return DROP;
	}

	/*N(rt->rt_refcnt--;) */

	/* FIXME: set csum_flags to 0 ? */
	F(RT_UNLOCK(rt);)

fwd:
	if(unlikely(((rt->rt_flags & RTF_GATEWAY) 
	N(|| ((rt->rt_ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)))) )) {
		/* icmp or redirect to backup_server
		icmp_error(mb, ICMP_UNREACH, ICMP_UNREACH_HOST, 0, 0); */
		srv->err_unreach++;

		if(!((srv->err_unreach - 1) % LIMIT_LOG))
			log(LOG_ERR, "drlb: Host %s unreachable !\n", paddr(srv->ip));

		STOP(fwd_mb);
		return DROP;
	}

	rt->rt_rmx.rmx_pksent++;

	(*rt->rt_ifp->if_output)(rt->rt_ifp, mb, (struct sockaddr *)srv->ip, rt);

	STOP(fwd_mb);
	return 0;
}




static void send_cnx(struct sockaddr_storage *src, struct sockaddr_storage *dst, struct sockaddr_storage *srv)
{
	struct mbuf *m;
	struct udpiphdr *ui;
	struct ip *ip;
	int len;
	struct sockaddr_in *s4_src = NULL;
	struct sockaddr_in *s4_dst = NULL;
	struct sockaddr_in6 *s6_src = NULL;
	struct sockaddr_in6 *s6_dst = NULL;

	START(snd_cnx);

	if(brother.ip->ss_family == AF_INET) {
		s4_dst = (struct sockaddr_in *)(brother.ip);
		if(brother.rt->rt_ifa) s4_src = IA_SIN(brother.rt->rt_ifa);
	} else if(brother.ip->ss_family == AF_INET6) {
		s6_dst = (struct sockaddr_in6 *)(brother.ip);
		if(brother.rt->rt_ifa) s6_src = (struct sockaddr_in6 *)(brother.rt->rt_ifa);
	} else {
		STOP(snd_cnx);
		return;
	}


	if(srv)
		len = 3 * sizeof(struct sockaddr_storage) + sizeof(uint32_t);
	else 
		len = 2 * sizeof(struct sockaddr_storage) + sizeof(uint32_t);

	m = m_gethdr(M_DONTWAIT, MT_DATA);

	if (m == NULL) {
		stats_nombuf++;
		if(!((stats_nombuf - 1) % LIMIT_LOG)) log(LOG_ERR, "drlb: can't allocate mbuf!\n");
		STOP(snd_cnx);
		return;
	}

	M_PREPEND(m, sizeof(struct udpiphdr), M_DONTWAIT);

	/* TODO : make same for IPv6 */
	ui = mtod(m, struct udpiphdr *);
	ui->ui_pr = IPPROTO_UDP;
	ui->ui_src = s4_src->sin_addr;
	ui->ui_dst = s4_dst->sin_addr;
	ui->ui_sport = sync_port;
	ui->ui_dport = s4_dst->sin_port;
	ui->ui_ulen = htons((u_int16_t)len + sizeof(struct udphdr));

	ui->ui_sum = in_pseudo(ui->ui_src.s_addr, s4_dst->sin_addr.s_addr, htons((u_short)len + sizeof(struct udphdr) + IPPROTO_UDP));
	m->m_pkthdr.csum_flags = CSUM_UDP;
	m->m_pkthdr.csum_data = offsetof(struct udphdr, uh_sum);

	ip = (struct ip *)&ui->ui_i;
	ip->ip_v = IPVERSION;
	ip->ip_hl = 5;
	ip->ip_len = htons(sizeof(struct udpiphdr) + len);
	ip->ip_tos = 0;
	ip->ip_off = 0;
	ip->ip_ttl = 32;
	ip->ip_id = ip_newid();
	ip->ip_sum = 0;
	ip->ip_sum = in_cksum(m, 5 << 2);


	m_copyback(m, sizeof(struct udpiphdr), sizeof(uint32_t), (caddr_t)&sync_key);
	m_copyback(m, sizeof(struct udpiphdr) + sizeof(uint32_t), sizeof(struct sockaddr_storage), (caddr_t)src);
	m_copyback(m, sizeof(struct udpiphdr) + sizeof(uint32_t) + sizeof(struct sockaddr_storage), sizeof(struct sockaddr_storage), (caddr_t)dst);
	if(srv)
		m_copyback(m, sizeof(struct udpiphdr) + sizeof(uint32_t) + 2 * sizeof(struct sockaddr_storage), sizeof(struct sockaddr_storage), (caddr_t)srv);

	/*
	D(log(LOG_DEBUG, "Send: %s.%i -> ", paddr(src), ntohs(SPORT(src)));)
	D(log(LOG_DEBUG, "%s.%i", paddr(dst), ntohs(SPORT(dst)));)
	D(log(LOG_DEBUG, " => %s.%i\n", srv ? paddr(srv) : "drop", srv ? ntohs(SPORT(srv)) : 0);)
	*/

	fwd(&brother, m);

	STOP(snd_cnx);
}


/* CALLOUT: on drop une cnx inactive */
static void drop_cnx(void *a)
{
	struct lbdr_conn *cnx = (struct lbdr_conn *)a;
	struct lbdr_conn *tmp;
	struct lbdr_conn *last = NULL;
	IF_NOBIG(kmutex_t *mtx;)
	int seen;

	IF_NOBIG(mtx = cnx->pmtx;)
	MTX_LOCK();

	START(dr_cnx);

	if (callout_pending(&(cnx->cal))) {
		D(log(LOG_DEBUG, "drop_cnx(): pending\n");)
		goto unlock;
	}
	if (!callout_active(&(cnx->cal))) {
		D(log(LOG_DEBUG, "drop_cnx(): not active\n");)
		goto unlock;
	}
	callout_deactivate(&(cnx->cal));

	if(cnx->status == CNX_SHARED) {
		cnx->srv->shared_cnx--;
	} else {

		if(cnx->status & CNX_FLAG_DATA)
			send_cnx(&(cnx->ip_src), cnx->parent->ip, NULL);

		if(cnx->status & CNX_FLAG_ACT) {
			stats_dropact++;
			if(!((stats_dropact - 1) % LIMIT_LOG))
				log(LOG_ERR, "drlb: drop_cnx() cnx already active!\n");
			/* cnx->srv->no_fin++; */
			cnx->srv->active_cnx--;
		} else {
			if(cnx->status != CNX_CLOSE)
				cnx->srv->no_fin++;
			cnx->srv->inactive_cnx--;
		}
	}

	seen = cnx->last_seen - time_second + (cnx->parent->to->t_clo/nticks);

	if(seen > 2)
		log(LOG_DEBUG, "drlb: warning %s.%i (was %s) last seen %d sec after close\n", paddr(&(cnx->ip_src)), ntohs(SPORT((&(cnx->ip_src)))), pstatus(cnx->status), seen);

	DTR(else log(LOG_DEBUG, "[Del] %s.%i is dropped (was %s) last seen %d sec after close\n", paddr(&(cnx->ip_src)), ntohs(SPORT((&(cnx->ip_src)))), pstatus(cnx->status), seen);)

	tmp = &(cnx->parent->htable[cnx->hash]);

	if(cnx == tmp) {
		cnx->srv = NULL;
		cnx->parent = NULL;
		cnx->hash = 0;
		cnx->status = 0;
		bzero(&(cnx->ip_src), sizeof(struct sockaddr_storage));
		/*N(callout_destroy(&(cnx->cal));)*/
		goto unlock;
	} else {
		while(tmp->next) {
			last = tmp;
			tmp = tmp->next;
			if(tmp == cnx) {
				last->next = cnx->next;
				N(callout_destroy(&(cnx->cal));)
				free(cnx, M_BUF);
				goto unlock;
			}
		}
	}

	log(LOG_ERR, "drlb: drop_cnx() cnx not found !\n");
unlock:
	STOP(dr_cnx);
	MTX_UNLOCK();
}





/* CALLOUT: on passe la cnx en inactive */
static void close_cnx(void *a)
{
	struct lbdr_conn *cnx = (struct lbdr_conn *)a;
	IF_NOBIG(kmutex_t *mtx;)

	IF_NOBIG(mtx = cnx->pmtx;)
	MTX_LOCK();

	START(cl_cnx);

	if (callout_pending(&(cnx->cal))) {
		D(log(LOG_DEBUG, "close_cnx(): pending\n");)
		goto unlock;
	}
	if (!callout_active(&(cnx->cal))) {
		D(log(LOG_DEBUG, "close_cnx(): not active\n");)
		goto unlock;
	}
	callout_deactivate(&(cnx->cal));

	/* FIXME: must be written ?? */
	D(log(LOG_DEBUG, "drlb: [Del] from %s.%i on timeout (was %s)\n", paddr(&(cnx->ip_src)), ntohs(SPORT((&(cnx->ip_src)))), pstatus(cnx->status));)


	if(likely(cnx->status & CNX_FLAG_ACT)) {
		cnx->srv->active_cnx--;
		cnx->srv->inactive_cnx++;
		cnx->status &= ~CNX_FLAG_ACT;
	} else {
		log(LOG_ERR, "drlb: close_cnx() closing already inactive cnx!\n");
	}

	callout_reset(&(cnx->cal), cnx->parent->to->t_clo, drop_cnx, (void *)cnx);

unlock:
	STOP(cl_cnx);
	MTX_UNLOCK();
}



static __inline__ int new_cnx(struct lbdr_service *service, struct lbdr_conn **c, u_char thflags, struct sockaddr_storage *ip, uint32_t h)
{
	struct lbdr_conn *cnx;
	uint32_t srv;

	START(nw_cnx);

	if(!(thflags & TH_SYN)) {
		service->to->err_nosyn++;
		STOP(nw_cnx);
		return NOTHING;
	}

	srv = (*service->sched)(service->server, service->nb_server, service->used_server);

	if(unlikely(srv == -1)) {
		service->to->err_nosrv++;
		if(!((service->to->err_nosrv - 1) % LIMIT_LOG))
			log(LOG_ERR, "drlb: No server found!\n");
		STOP(nw_cnx);
		return NOTHING;
	}

	service->used_server = srv;

	cnx = create_cnx(service, h, srv, ip);

	if(unlikely(!cnx)) {
		stats_nomem++;
		if(!((stats_nomem - 1) % LIMIT_LOG))
			log(LOG_ERR, "drlb: not enough kernel memory!\n");
		STOP(nw_cnx);
		return DROP;
	}
	
	cnx->srv->nb_cnx++;
	cnx->srv->active_cnx++;
	callout_reset(&(cnx->cal), service->to->t_syn, close_cnx, (void *)cnx);
	*c = cnx;


	STOP(nw_cnx);
	return 0;
}



static int recv_cnx(struct sockaddr_storage *src, struct sockaddr_storage *dst, struct sockaddr_storage *srv)
{
	struct lbdr_conn *cnx;
	struct lbdr_service *service;
	IF_NOBIG(kmutex_t *mtx;)
	uint32_t h;
	int server;

	START(rcv_cnx);

	service = find_service(dst);

	if(!service) {
		TRACE("%s.%d not configured, dropped\n", paddr(dst), ntohs(SPORT(dst)));
		STOP(rcv_cnx);
		return -1;
	}

	TRACE("for %s.%d, ", paddr(dst), ntohs(SPORT(dst)));
	TRACE("from %s.%d, ", paddr(src), ntohs(SPORT(src)));

	service->refcnt++;

	if(src->ss_family == AF_INET6)
		h = sfhash6(src) & (service->htsize-1);
	else
		h = sfhash4(src) & (service->htsize-1);

	IF_NOBIG(mtx = &(service->locktable[h]);)
	MTX_LOCK();

	cnx = find_cnx(service->htable, h, src);

	if(cnx) {
		TRACE("removed\n");
		callout_reset(&(cnx->cal), 1, drop_cnx, (void *)cnx);
		goto unlock;
	}

	if(!srv->ss_family) {
		TRACE("not registered, dropped\n");
		goto unlock;
	}

	server = find_server(service->server, service->nb_server, srv);

	if(unlikely(server == -1)) {
		stats_notsame++;
		if(!((stats_notsame - 1) % LIMIT_LOG)) log(LOG_ERR, "drlb: server %s not know.\n", paddr(srv));
		TRACE("unknow server, dropped\n");
		goto unlock;
	}

	cnx = create_cnx(service, h, server, src); 

	if(unlikely(!cnx)) {
		stats_nomem++;
		if(!((stats_nomem - 1) % LIMIT_LOG)) log(LOG_ERR, "drlb: not enough kernel memory!\n");
		TRACE("can't create cnx, dropped\n");
		goto unlock;
	}

	/* FIXME: add if_xname */
	ADDLOG("drlb: [Rmt] cnx from %s.%d ", paddr(src), ntohs(SPORT(src)));
	ADDLOG("to %s.%d ", paddr(dst), ntohs(SPORT(dst)));
	ADDLOG("fwd to %s.%d ", paddr(srv), ntohs(SPORT(srv)));

	TRACE("server %s, registered\n", paddr(srv));

	cnx->status = CNX_SHARED;
	cnx->srv->shared_cnx++;
	callout_reset(&(cnx->cal), 86400 * nticks, drop_cnx, (void *)cnx); /* keep it max 24 hours if no response ? */

unlock:
	MTX_UNLOCK();
	service->refcnt--;
	STOP(rcv_cnx);

	return 0;
}


/**
 * change cnx status
 */
static __inline__ int update_status_cnx(struct lbdr_service *service, struct lbdr_conn *cnx, u_char thflags)
{
	uint32_t srv;

	START(up_cnx);


	if(thflags & TH_SYN) {

		TRACE("?S");
		if(cnx->status & CNX_FLAG_ACT) {
			cnx->srv->active_cnx--;
		} else {
			cnx->srv->inactive_cnx--;
		}

		srv = (*service->sched)(service->server, service->nb_server, service->used_server);

		if(srv == -1) {
			service->to->err_nosrv++;
		} else {
			service->used_server = srv;
			cnx->srv = service->server[srv];
		}

		cnx->srv->nb_cnx++;
		cnx->srv->active_cnx++;
		cnx->status = CNX_SYN;

		callout_reset(&(cnx->cal), service->to->t_syn, close_cnx, (void *)cnx);

	} else if(thflags & TH_FIN || thflags & TH_RST) {

		TRACE("%c ", thflags & TH_FIN ? 'F' : 'R');

		if(cnx->status & CNX_FLAG_ACT) {
			cnx->srv->active_cnx--;
			cnx->srv->inactive_cnx++;
		}

		D(log(LOG_DEBUG, "drlb: [Del] from %s.%i with %s (was %s)\n", paddr(&(cnx->ip_src)), ntohs(SPORT((&(cnx->ip_src)))), thflags & TH_FIN ? "FIN" : "RST", pstatus(cnx->status));)

		/* on ne partage pas les cnx qui ont recu juste SYN puis RST */
		if(cnx->status & CNX_FLAG_DATA)
			cnx->status = CNX_CLOSE;
		else
			cnx->status = CNX_SYNTO;


		callout_reset(&(cnx->cal), cnx->parent->to->t_clo, drop_cnx, (void *)cnx);

	} else if(cnx->status == CNX_ESTABL) {
			callout_sched(&(cnx->cal), service->to->t_est, close_cnx, (void *)cnx);
	} else if(cnx->status == CNX_SYN) {
			send_cnx(&(cnx->ip_src), service->ip, cnx->srv->ip);
			cnx->status = CNX_ESTABL;
			callout_sched(&(cnx->cal), service->to->t_est, close_cnx, (void *)cnx);
	} else if(cnx->status == CNX_TIMEOUT || cnx->status == CNX_SYNTO) {
			log(LOG_WARNING, "drlb: warning %s.%i (was %s) not seen for %d sec\n",
					paddr(&(cnx->ip_src)), ntohs(SPORT((&(cnx->ip_src)))),
					pstatus(cnx->status), (int)time_second - cnx->last_seen);
			cnx->srv->inactive_cnx--;
			cnx->srv->active_cnx++;
			cnx->status = CNX_ESTABL;
			callout_reset(&(cnx->cal), service->to->t_est, close_cnx, (void *)cnx);
	} else if(cnx->status == CNX_SHARED) {
			cnx->srv->active_cnx++;
			cnx->srv->shared_cnx--;
			cnx->status = CNX_ESTABL;
			callout_reset(&(cnx->cal), service->to->t_est, close_cnx, (void *)cnx);
	}

	cnx->last_seen = time_second;

	STOP(up_cnx);

	return 0;
}


static __inline__ int update_status_ssl_cnx(struct lbdr_service *service, struct lbdr_conn *cnx, u_char thflags)
{
	START(up_cnx);

	if(!(cnx->status & CNX_FLAG_ACT)) {
		cnx->srv->active_cnx++;
		cnx->srv->inactive_cnx--;
	}

	cnx->status = CNX_ESTABL;
	callout_reset(&(cnx->cal), service->to->t_est, close_cnx, (void *)cnx);

	STOP(up_cnx);

	return 0;
}

static int incoming_icmp4(struct mbuf **m, struct ip *iph)
{
	struct mbuf *mb = *m;
	struct sockaddr_storage ssrc, sdst;
	struct lbdr_service *service;
	struct lbdr_conn *cnx;
	struct sockaddr_in *in_src = (struct sockaddr_in *)&ssrc;
	struct sockaddr_in *in_dst = (struct sockaddr_in *)&sdst;
	struct icmp *ich;
	IF_NOBIG(kmutex_t *mtx;)
	uint32_t h;

	if (mb->m_len < (sizeof (struct ip) + sizeof(struct icmp))) {
		if ((mb = m_pullup(mb, sizeof (struct ip) + sizeof(struct icmp))) == NULL) {
			return 0;
		}
	}

	ich = (struct icmp *)((char *)iph + (iph->ip_hl << 2));

	switch(ich->icmp_type) {
		case ICMP_UNREACH:
		case ICMP_REDIRECT:
		case ICMP_SOURCEQUENCH:
		case ICMP_TIMXCEED:
		case ICMP_PARAMPROB:
			break;
		default:
			DG(log(LOG_DEBUG, "ICMP: %s ignored\n", picmptype(ich->icmp_type));)
			return 0;
	}

	in_src->sin_family = AF_INET;
	in_src->sin_len = sizeof(struct sockaddr_in);
	in_src->sin_port = 0;
	in_src->sin_addr = ich->icmp_ip.ip_src;

	in_dst->sin_family = AF_INET;
	in_dst->sin_len = sizeof(struct sockaddr_in);
	in_dst->sin_port = 0;
	in_dst->sin_addr = ich->icmp_ip.ip_dst;

	TRACE("ICMP %s src-ip: %s ", picmptype(ich->icmp_type), paddr(&ssrc));
	TRACE("dst-ip: %s ", paddr(&sdst));

	service = find_service(&ssrc);

	if(!service) {
		TRACE("doesn't correspond to registered service, ignored\n");
		PRINT_TRACE();
		return 0;
	}

	service->refcnt++;

	h = sfhash4(&sdst) & (service->htsize-1);

	IF_NOBIG(mtx = &(service->locktable[h]);)

	MTX_LOCK();

	cnx = find_cnx(service->htable, h, &sdst);

	if(cnx) {
		TRACE("owned by existing cnx, forwarded\n");
		ADDLOG("drlb: icmp from %s.%d ", paddr(&sdst), ntohs(SPORT((&sdst))));
		ADDLOG("sent to %s.%d\n", paddr(&ssrc), ntohs(SPORT((&ssrc))));
		fwd(cnx->srv, mb);
	} else {
		TRACE("doesn't correspond to existing cnx, dropped\n");
	}

	MTX_UNLOCK();

	service->refcnt--;

	m_freem(*m);
	*m = NULL;

	PRINT_TRACE();
	PRINT_LOG();

	return 1;
}

/* TODO udp6 */
static __inline__ int incoming_udp4(struct mbuf **m, struct ip *iph)
{
	struct sockaddr_storage ssrc, src, dst, srv;
	struct sockaddr_in *sinsrc = (struct sockaddr_in *)&ssrc;
	//struct sockaddr_in *sindst = (struct sockaddr_in *)&sdst;
	struct mbuf *mb = *m;
	struct udphdr *uh;
	uint32_t h;
	int len;
	u_short uh_sum;

	START(inc_udp);

	if(!brother.ip->ss_family) {
		STOP(inc_udp);
		return 0;
	}


	/* and iphlen ? */
	if (mb->m_len < sizeof (struct udpiphdr)) {
		if ((mb = m_pullup(mb, sizeof (struct udpiphdr))) == NULL) {
			STOP(inc_udp);
			return 0;
		}
	}

	uh = (struct udphdr *)((char *)iph + (iph->ip_hl << 2));
	len = ntohs(uh->uh_ulen);

	sinsrc->sin_len = sizeof(struct sockaddr_in);
	sinsrc->sin_family = AF_INET;
	sinsrc->sin_port = uh->uh_sport;
	sinsrc->sin_addr = iph->ip_src;

	TRACE("UDP src-ip: %s ", paddr(&ssrc));

	if(len != (mb->m_len - sizeof(struct ip))) {
		/* Is that util ? */
		TRACE("bad udp-len, ignored\n");
		PRINT_TRACE();
		STOP(inc_udp);
		return 0;
	}

	if(same_addr(brother.ip, &ssrc)
		F(&& (mb->m_pkthdr.csum_flags & CSUM_DATA_VALID))
		) {

		/* test my addr/port:
		sindst->sin_len = sizeof(struct sockaddr_in);
		sindst->sin_family = AF_INET;
		sindst->sin_port = uh->uh_dport;
		sindst->sin_addr = iph->ip_dst;
		*/

#ifdef __FreeBSD__
		if (mb->m_pkthdr.csum_flags & CSUM_PSEUDO_HDR)
			uh_sum = mb->m_pkthdr.csum_data;
		else
#endif
			uh_sum = in_pseudo(iph->ip_src.s_addr, iph->ip_dst.s_addr, htonl((u_short)len + mb->m_pkthdr.csum_data + IPPROTO_UDP));

		uh_sum ^= 0xffff;

		if (uh_sum) {
			TRACE("bad udp-csum, dropped\n");
			stats_udpbadcsum++;
			goto dropit;
		}
		
		m_copydata(mb, sizeof(struct udpiphdr), sizeof(uint32_t), (caddr_t)&h);

		if(h != sync_key) {
			TRACE("bad sync-key, dropped\n");
			goto dropit;
		}

		bzero(&srv, sizeof(struct sockaddr_storage));

		switch(len) {
		case (3*sizeof(struct sockaddr_storage) + sizeof(struct udphdr) + sizeof(uint32_t)):
			m_copydata(mb, sizeof(struct udpiphdr) + sizeof(uint32_t) + 2 * sizeof(struct sockaddr_storage), sizeof(struct sockaddr_storage), (caddr_t)&srv);
		case (2*sizeof(struct sockaddr_storage)+sizeof(struct udphdr) + sizeof(uint32_t)):
			m_copydata(mb, sizeof(struct udpiphdr) + sizeof(uint32_t) + sizeof(struct sockaddr_storage), sizeof(struct sockaddr_storage), (caddr_t)&dst);
			m_copydata(mb, sizeof(struct udpiphdr) + sizeof(uint32_t), sizeof(struct sockaddr_storage), (caddr_t)&src);
			break;
		default:
			TRACE("bad data-len, dropped\n");
			goto dropit;
		}

		recv_cnx(&src, &dst, &srv);
	dropit:
		m_freem(*m);
		*m = NULL;
	}

	PRINT_TRACE();
	PRINT_LOG();

	STOP(inc_udp);

	return 0;
}


/*
static __inline__ int incoming_udp6(struct mbuf **m, struct ip6_hdr *iph6)
{
	struct sockaddr_storage ssrc, src, dst, srv;
	struct sockaddr_in6 *sin6src = (struct sockaddr_in6 *)&ssrc;
	//struct sockaddr_in6 *sin6dst = (struct sockaddr_in6 *)&sdst;
	struct mbuf *mb = *m;
	struct udphdr *uh;
	uint32_t h;
	int len;
	u_short uh_sum;

	START(inc_udp);

	if(!brother.ip->ss_family) {
		STOP(inc_udp);
		return 0;
	}

	if (mb->m_len < (sizeof (struct udphdr) + sizeof(struct ip6_hdr))) {
		if ((mb = m_pullup(mb, (sizeof (struct udphdr) + sizeof(struct ip6_hdr)))) == NULL) {
			STOP(inc_udp);
			return 0;
		}
	}

        uh = (struct udphdr *)iph6 + sizeof(struct ip6_hdr);
	len = ntohs(uh->uh_ulen);

	sin6src->sin6_len = sizeof(struct sockaddr_in6);
	sin6src->sin6_family = AF_INET6;
	sin6src->sin6_port = uh->uh_sport;
	sin6src->sin6_addr = iph6->ip6_src;

	TRACE("UDP src-ip: %s ", paddr(&ssrc));

	if(len != (mb->m_len - sizeof(struct ip))) {
		TRACE("bad udp-len, ignored\n");
		PRINT_TRACE();
		STOP(inc_udp);
		return 0;
	}

	if(same_addr(brother.ip, &ssrc) && (mb->m_pkthdr.csum_flags & CSUM_DATA_VALID)) {

		if (uh->uh_sum == 0 || in6_cksum(mb, IPPROTO_UDP, off, ulen) != 0) {
			TRACE("bad udp-csum, dropped\n");
			stats_udpbadcsum++;
			goto dropit;
		}
		
		m_copydata(mb, sizeof(struct udpiphdr), sizeof(uint32_t), (caddr_t)&h);

		if(h != sync_key) {
			TRACE("bad sync-key, dropped\n");
			goto dropit;
		}

		bzero(&srv, sizeof(struct sockaddr_storage));

		switch(len) {
		case (3*sizeof(struct sockaddr_storage) + sizeof(struct udphdr) + sizeof(uint32_t)):
			m_copydata(mb, sizeof(struct udpiphdr) + sizeof(uint32_t) + 2 * sizeof(struct sockaddr_storage), sizeof(struct sockaddr_storage), (caddr_t)&srv);
		case (2*sizeof(struct sockaddr_storage)+sizeof(struct udphdr) + sizeof(uint32_t)):
			m_copydata(mb, sizeof(struct udpiphdr) + sizeof(uint32_t) + sizeof(struct sockaddr_storage), sizeof(struct sockaddr_storage), (caddr_t)&dst);
			m_copydata(mb, sizeof(struct udpiphdr) + sizeof(uint32_t), sizeof(struct sockaddr_storage), (caddr_t)&src);
			break;
		default:
			TRACE("bad data-len, dropped\n");
			goto dropit;
		}

		recv_cnx(&src, &dst, &srv);
	dropit:
		m_freem(*m);
		*m = NULL;
	}

	PRINT_TRACE();
	PRINT_LOG();

	STOP(inc_udp);

	return 0;
}
*/

/***
 * main functions (pfil entry)
 */


/**
 * PFIL input for IPv4
 * 
 * FIXME: doesn't take care of options field in IP header (for m_pullup())!
 *
 */
static int
F(lb4_input(void *arg, struct mbuf **m, struct ifnet *i, int dir, struct inpcb *inp))
N(lb4_input(void *arg, struct mbuf **m, struct ifnet *i, int dir))
{
        struct mbuf *mb;
        struct ip *iph;
	struct sockaddr_storage ss;
	struct sockaddr_in *in = (struct sockaddr_in *)&ss;
	struct lbdr_service *service;
	struct lbdr_conn *cnx;
	struct tcphdr *th;
	IF_NOBIG(kmutex_t *mtx;)
	uint32_t hash;

	mb = *m;
	iph = mtod(mb, struct ip *);

	switch(iph->ip_p) {
		case IPPROTO_UDP:
			/* FIXME xen: */
			/*return incoming_udp4(m, iph);*/
			return 0;
		case IPPROTO_TCP:
			break;
		case IPPROTO_ICMP:
			return incoming_icmp4(m, iph);
		default:
			return 0;
	}

	if (mb->m_len < sizeof (struct ip) + sizeof(struct tcphdr)) {
		if ((mb = m_pullup(mb, sizeof (struct ip) + sizeof(struct tcphdr))) == NULL) {
			DG(log(LOG_DEBUG, "mbuf too short\n");)
			stats_badmbuf++;
			return 0;
		}
	}

        th = (struct tcphdr *)((char *)iph + (iph->ip_hl << 2));

	in->sin_len = sizeof(struct sockaddr_in);
	in->sin_family = AF_INET;
	in->sin_port = th->th_dport;
	in->sin_addr = iph->ip_dst;

	service = find_service4(&ss);

	if(!service) return 0;

	service->refcnt++;

	START(input);

        D({
		RESET_LOG();
		in->sin_port = th->th_sport;
		in->sin_addr = iph->ip_src;
        	TRACE("%s %s:%s:%i -> ", i->if_xname, pproto(iph->ip_p), paddr((struct sockaddr_storage *)in), ntohs(in->sin_port));
        	ADDLOG("drlb: [New] from %s:%s.%i ", i->if_xname, paddr((struct sockaddr_storage *)in), ntohs(in->sin_port));
		in->sin_port = th->th_dport;
		in->sin_addr = iph->ip_dst;
        	TRACE("%s:%i ", paddr((struct sockaddr_storage *)in), ntohs(in->sin_port));
        	ADDLOG("to %s:%i ", paddr((struct sockaddr_storage *)in), ntohs(in->sin_port));
	})


	in->sin_addr = iph->ip_src;
	in->sin_port = service->persist ? 0 : th->th_sport;

	hash = sfhash4((struct sockaddr_storage *)in) & (service->htsize-1);

	IF_NOBIG(mtx = &(service->locktable[hash]);)
	MTX_LOCK();

	cnx = find_cnx(service->htable, hash, (struct sockaddr_storage *)in);

	if(!cnx) {
		TRACE("%s ", th->th_flags & TH_SYN ? "S" : "!S");
		switch(new_cnx(service, &cnx, th->th_flags, (struct sockaddr_storage *)in, hash)) {
			case DROP:
				goto drop;
			case NOTHING:
				goto nothing;
			default:
				ADDLOG("fwd to %s:%s.%d\n", cnx->srv->rt->rt_ifp->if_xname, paddr(cnx->srv->ip), ntohs(SPORT(cnx->srv->ip)));
				PRINT_LOG();
				break;
		}
		
	} else { 
		if(unlikely(service->persist))
			update_status_ssl_cnx(service, cnx, th->th_flags);
		else
			update_status_cnx(service, cnx, th->th_flags);
	}

	/* only needed on FreeBSD :-) ! */
	F(iph->ip_len = htons(iph->ip_len);)
	F(iph->ip_off = htons(iph->ip_off);)

	if(fwd(cnx->srv, mb) == DROP) goto drop;
	DTR(else TRACE("=> %s.%i ", paddr(cnx->srv->ip), ntohs(SPORT(cnx->srv->ip)));)

        cnx->srv->pkts++;
        cnx->srv->bytes += mb->m_len;

        *m = NULL;

nothing:
	TRACE("\n");
	PRINT_TRACE();

	service->refcnt--;
	MTX_UNLOCK();
	STOP(input);

	return 0;

drop:
	TRACE("DROP\n");
	PRINT_TRACE();

	m_freem(*m);
        *m = NULL;
	service->refcnt--;
	MTX_UNLOCK();
	STOP(input);

	return 1;
}


/**
 * PFIL entry for IPv6
 */
static int
F(lb6_input(void *arg, struct mbuf **m, struct ifnet *i, int dir , struct inpcb *inp))
N(lb6_input(void *arg, struct mbuf **m, struct ifnet *i, int dir))
{
        struct mbuf *mb;
        struct ip6_hdr *iph6;
	struct sockaddr_storage ss;
	struct sockaddr_in6 *in6;
	struct lbdr_service *service;
	struct lbdr_conn *cnx;
	struct tcphdr *th;
	IF_NOBIG(kmutex_t *mtx;)
	uint32_t hash;

	mb = *m;
	iph6 = mtod(mb, struct ip6_hdr *);

        switch(iph6->ip6_nxt) {
        	case IPPROTO_TCP:
        		break;
        	case IPPROTO_UDP:
        		/*incoming_udp6();*/
        	case IPPROTO_ICMP:
        	default:
        		return 0;
        }

	if (mb->m_len < sizeof (struct ip6_hdr) + sizeof(struct tcphdr)) {
		if ((mb = m_pullup(mb, sizeof (struct ip6_hdr) + sizeof(struct tcphdr))) == NULL) {
			DG(log(LOG_DEBUG, "mbuf too short");)
			stats_badmbuf++;
			return 0;
		}
	}

        /* Add ipv6 options !!! */
        th = (struct tcphdr *)iph6 + sizeof(struct ip6_hdr);

	in6 = (struct sockaddr_in6 *)&ss;
	in6->sin6_len = sizeof(struct sockaddr_in6);
	in6->sin6_family = AF_INET6;
	in6->sin6_port = th->th_dport;
	in6->sin6_addr = iph6->ip6_dst;

	service = find_service6((struct sockaddr_storage *)in6);

	if(!service) return 0;

	service->refcnt++;

        D({
		RESET_LOG();
		in6->sin6_port = th->th_sport;
		in6->sin6_addr = iph6->ip6_src;
        	TRACE("%s %s/%s.%d -> ", i->if_xname, pproto(iph6->ip6_nxt), paddr((struct sockaddr_storage *)in6), ntohs(in6->sin6_port));
        	ADDLOG("drlb: [New] cnx from %s/%s.%d ", i->if_xname, paddr((struct sockaddr_storage *)in6), ntohs(in6->sin6_port));
		in6->sin6_port = th->th_dport;
		in6->sin6_addr = iph6->ip6_dst;
        	TRACE("%s.%d ", paddr((struct sockaddr_storage *)in6), ntohs(in6->sin6_port));
        	ADDLOG("for %s.%d  ", paddr((struct sockaddr_storage *)in6), ntohs(in6->sin6_port));
	})

	in6->sin6_addr = iph6->ip6_src;
	in6->sin6_port = service->persist ? 0 : th->th_sport;

	hash = sfhash6((struct sockaddr_storage *)in6) & (service->htsize-1);

	IF_NOBIG(mtx = &(service->locktable[hash]);)
	MTX_LOCK();

	cnx = find_cnx(service->htable, hash, (struct sockaddr_storage *)in6);

	if(!cnx) {
		TRACE("%s ", th->th_flags & TH_SYN ? "S" : "!S");
		switch(new_cnx(service, &cnx, th->th_flags, (struct sockaddr_storage *)in6, hash)) {
			case DROP:
				goto drop;
			case NOTHING:
				goto nothing;
			default:
				ADDLOG("fwd to %s.%d\n", paddr(cnx->srv->ip), ntohs(SPORT(cnx->srv->ip)));
				PRINT_LOG();
				break;
		}
		
	} else { 
		if(service->persist)
			update_status_ssl_cnx(service, cnx, th->th_flags);
		else
			update_status_cnx(service, cnx, th->th_flags);
	}

	if(fwd(cnx->srv, mb) == NOTHING) goto drop;
	DTR(else TRACE("=> %s.%i ", paddr(cnx->srv->ip), ntohs(SPORT(cnx->srv->ip)));)

        cnx->srv->pkts++;
        cnx->srv->bytes += mb->m_len;

        *m = NULL;

nothing:
	TRACE("\n");
	PRINT_TRACE();

	service->refcnt--;
	MTX_UNLOCK();


	return 0;

drop:
	TRACE("DROP\n");
	PRINT_TRACE();

	m_freem(*m);
        *m = NULL;
	service->refcnt--;
	MTX_UNLOCK();
	return 1;
}


/***
 * config and statistique:
 *
 * dump_stats();
 * ioctl();
 */
static u_int get_nb_ip(unsigned long *nb_cnx)
{
	struct lbdr_service *s = NULL;
	struct lbdr_server **srv;
	int i, r = 0, j = 0;;

	for(i = 0; i < LIMIT_SERVICE; i++) {
		s = service_available[i];
		while(s) {
			r++;
			if(!s->cloned) {
				r += s->nb_server;
				srv = s->server;
				for(j = 0; j < s->nb_server; j++) {
					*nb_cnx += srv[j]->nb_cnx;
				}
			}
			s = s->next;
		}
	}

	/*
	D(uprintf("drlb: got %i IPs\n", r);)
	*/

	return r;
}


static int dump_stats(struct lbdr_param *param)
{
	struct lbdr_service *s = NULL;
	struct lbdr_server **srv = NULL;
	struct lbdr_stats ls, *dest;
	int i, j, max_out = 0, nb_out = 0;

	max_out = param->size / sizeof(struct lbdr_stats);

	/*
	D(uprintf("drlb: Can output %i IPs\n", max_out);)
	*/

	dest = param->data;

	for(i = 0; i < LIMIT_SERVICE; i++) {
		s = service_available[i];
		while(s) {
			if(!s->cloned) {
				bzero(&ls, sizeof(struct lbdr_stats));
				copy_addr(&(ls.ip), s->ip);

				ls.type = (i << 16 | LBDR_TYPE_SERVICE);
				ls.stats.svi.ls_sched = sched_number(s->sched);
				ls.stats.svi.ls_coll = s->to->nb_coll;
				ls.stats.svi.ls_persist = s->persist;
				ls.stats.svi.ls_err_nosyn = s->to->err_nosyn;
				ls.stats.svi.ls_err_nosrv = s->to->err_nosrv;
				ls.stats.svi.ls_timeout_syn = s->to->t_syn / nticks;
				ls.stats.svi.ls_timeout_est = s->to->t_est / nticks;
				ls.stats.svi.ls_timeout_clo = s->to->t_clo / nticks;
				ls.stats.svi.ls_hashtablesize = s->htsize;

				if(nb_out < max_out) {
					if(copyout(&ls, &(dest[nb_out]), sizeof(struct lbdr_stats)))
						return EFAULT;
					nb_out++;
				} else {
					D(uprintf("Not enough memory allocated. Can't dump all data!\n");)
					return EIO;
				}

				FOREACHCLONE(s,
					bzero(&ls, sizeof(struct lbdr_stats));
					copy_addr(&(ls.ip), clone->ip);

					ls.type = (i << 16 | LBDR_TYPE_CLONE);
					ls.stats.svi.ls_sched = sched_number(s->sched);
					ls.stats.svi.ls_coll = s->to->nb_coll;
					ls.stats.svi.ls_persist = s->persist;
					ls.stats.svi.ls_err_nosyn = s->to->err_nosyn;
					ls.stats.svi.ls_err_nosrv = s->to->err_nosrv;
					ls.stats.svi.ls_timeout_syn = s->to->t_syn / nticks;
					ls.stats.svi.ls_timeout_est = s->to->t_est / nticks;
					ls.stats.svi.ls_timeout_clo = s->to->t_clo / nticks;
					ls.stats.svi.ls_hashtablesize = s->htsize;

					if(nb_out < max_out) {
						if(copyout(&ls, &(dest[nb_out]), sizeof(struct lbdr_stats)))
							return EFAULT;
						nb_out++;
					} else {
						D(uprintf("Not enough memory allocated. Can't dump all data!\n");)
						return EIO;
					}
				)

				srv = s->server;
				for(j = 0; j < s->nb_server; j++) {
					bzero(&ls, sizeof(struct lbdr_stats));
					copy_addr(&(ls.ip), srv[j]->ip);

					ls.type = (i << 16 | LBDR_TYPE_SERVER);
					ls.stats.srv.ls_weight = srv[j]->weight;
					ls.stats.srv.ls_rate = (srv[j]->cnx_sec) >> 16;
					ls.stats.srv.ls_nb_cnx = srv[j]->nb_cnx;
					ls.stats.srv.ls_act_cnx = srv[j]->active_cnx;
					ls.stats.srv.ls_inact_cnx = srv[j]->inactive_cnx;
					ls.stats.srv.ls_shared_cnx = srv[j]->shared_cnx;
					ls.stats.srv.ls_status = srv[j]->status;
					ls.stats.srv.ls_pkts = srv[j]->pkts;
					ls.stats.srv.ls_bytes = srv[j]->bytes;
					ls.stats.srv.ls_no_fin = srv[j]->no_fin;
					ls.stats.srv.ls_err_unreach = srv[j]->err_unreach;
					ls.stats.srv.ls_err_noroute = srv[j]->err_noroute;
					if(nb_out < max_out) {
						if(copyout(&ls, &(dest[nb_out]), sizeof(struct lbdr_stats)))
							return EFAULT;
						nb_out++;
					} else {
						D(uprintf("Not enough memory allocated. Can't dump all data!\n");)
						return EIO;
					}
				}
			}
			s = s->next;
		}
	}

	return 0;
}

static void dump_time(void)
{
	struct lbdr_service *service;
	struct lbdr_server **server;
	int i, n;

#ifdef DEBUGPRED
        long int *s = &__start_predict_data;
        long int *e = &__stop_predict_data;
        long int *sl = &__start_predict_line;
        const char **sf = &__start_predict_file;

        while (s < e) {
                uprintf("%s:%ld: incorrect=%ld, correct=%ld%s\n", *sf, *sl, s[0], s[1], s[0] > s[1] ? " <==== WARNING" : "");
                ++sl; ++sf; s += 2;
        }
#endif
#ifdef DEBUGTIME
	uprintf("Average cycles used:\n"
		  "   find_service() (%9u calls) %8u cycl/call\n"
		  "   find_server()  (%9u calls) %8u cycl/call\n"
		  "   sfhash()       (%9u calls) %8u cycl/call\n"
		  "   find_cnx()     (%9u calls) %8u cycl/call\n"
		  "   update_cnx     (%9u calls) %8u cycl/call\n"
		  "   new_cnx()      (%9u calls) %8u cycl/call\n"
		  "   create_cnx()   (%9u calls) %8u cycl/call\n"
		  "   close_cnx()    (%9u calls) %8u cycl/call\n"
		  "   send_cnx()     (%9u calls) %8u cycl/call\n"
		  "   recv_cnx()     (%9u calls) %8u cycl/call\n"
		  "   drop_cnx()     (%9u calls) %8u cycl/call\n"
		  "   scheduler()    (%9u calls) %8u cycl/call\n"
		  "   incoming_udp() (%9u calls) %8u cycl/call\n"
		  "   forward()      (%9u calls) %8u cycl/call\n"
		  "   input()        (%9u calls) %8u cycl/call\n"
		,
		fd_svi.nb,  fd_svi.nb  > 0 ? (u_int)(fd_svi.ns / fd_svi.nb) : 0,
		fd_srv.nb,  fd_srv.nb  > 0 ? (u_int)(fd_srv.ns / fd_srv.nb) : 0,
		hsh.nb,     hsh.nb     > 0 ? (u_int)(hsh.ns / hsh.nb) : 0,
		fd_cnx.nb,  fd_cnx.nb  > 0 ? (u_int)(fd_cnx.ns / fd_cnx.nb) : 0,
		up_cnx.nb,  up_cnx.nb  > 0 ? (u_int)(up_cnx.ns / up_cnx.nb) : 0,
		nw_cnx.nb,  nw_cnx.nb  > 0 ? (u_int)(nw_cnx.ns / nw_cnx.nb) : 0,
		cr_cnx.nb,  cr_cnx.nb  > 0 ? (u_int)(cr_cnx.ns / cr_cnx.nb) : 0,
		cl_cnx.nb,  cl_cnx.nb  > 0 ? (u_int)(cl_cnx.ns / cl_cnx.nb) : 0,
		snd_cnx.nb, snd_cnx.nb > 0 ? (u_int)(snd_cnx.ns / snd_cnx.nb) : 0,
		rcv_cnx.nb, rcv_cnx.nb > 0 ? (u_int)(rcv_cnx.ns / rcv_cnx.nb) : 0,
		dr_cnx.nb,  dr_cnx.nb  > 0 ? (u_int)(dr_cnx.ns / dr_cnx.nb) : 0,
		schd.nb,    schd.nb    > 0 ? (u_int)(schd.ns / schd.nb) : 0,
		inc_udp.nb, inc_udp.nb > 0 ? (u_int)(inc_udp.ns / inc_udp.nb) : 0,
		fwd_mb.nb, fwd_mb.nb > 0 ? (u_int)(fwd_mb.ns / fwd_mb.nb) : 0,
		input.nb, input.nb > 0 ? (u_int)(input.ns / input.nb) : 0
		);
#endif

	for(i = 0; i < LIMIT_SERVICE; i++) {
		service = service_available[i];
		while(service) {
			for(n = 0; n < service->nb_server; n++) {
				server = service->server;
				if(server[n]->rt->rt_flags & RTF_UP)
					printf("%s : %p  Flags: %x If: %s\n", paddr(server[n]->ip), server[n]->rt,
					server[n]->rt->rt_flags, server[n]->rt->rt_ifp->if_xname);
				else {
					printf("%s : %p  Flags: %x After try:\n", paddr(server[n]->ip), server[n]->rt, server[n]->rt->rt_flags);
					server[n]->rt = rtalloc1((struct sockaddr *)server[n]->ip, 1, 0UL);
					printf("\t %s : %p  Flags: %i\n", paddr(server[n]->ip), server[n]->rt, server[n]->rt->rt_flags);
				}
			}
			service = service->next;
		}
	}

	{ 
		unsigned long long int a, b;
		/*
		int f = 0;
		, c, d, e;
		c = 42; d = 42;
		*/
		a = read_tsc();
		printf("A: %llu\n", a);
		/*
		for (f = 0; f < 20; f++) {
			read_tsc();
		}
		b = read_tsc();
		printf("B: %lu\n", b);
		c = b - a;
		printf("C: %lu\n", c);
		*/

		a = read_tsc();
		/*printf("A: %lu\n", a);*/
		b = read_tsc();
		printf("READTSC: %llu\n", (b - a) / 5);
		b = read_tsc();
		/*printf("B: %lu\n", b);*/
		printf("READTSC: %llu\n", (b - a) / 5);
		b = read_tsc();
		/*printf("B: %lu\n", b);*/
		printf("READTSC: %llu\n", (b - a) / 5);
	}

}

/* 
 * Device functions
 */


static
N(int drlbioctl(dev_t d, u_long cmd, void *arg, int mode, struct lwp *lwp))
F(int drlbioctl(struct cdev *dev, u_long cmd, caddr_t arg, int mode, struct thread *td))
{
	int error = 0;
	struct lbdr_service * service, *last;
	struct lbdr_server **newsrv, **oldsrv;
	struct lbdr_conf *conf = (struct lbdr_conf *)arg;
	struct sockaddr_storage *dst = &(conf->ip_dst);
	int i;

        switch(cmd) {
        case LBDRADDCONF:
		DG(log(LOG_DEBUG, "LBDRADDSRV\n"));
		if(dst->ss_family == AF_INET6)
			((struct sockaddr_in6 *)&(conf->ip_srv))->sin6_port = SPORT(dst);
		else if(dst->ss_family == AF_INET)
			((struct sockaddr_in *)&(conf->ip_srv))->sin_port = SPORT(dst);

                service = register_service(dst, conf->sched, conf->persist, conf->hashtablesize);
                if(!service) {
                	error = ENOMEM;
                	break;
                }
		if(service->cloned) {
			D(uprintf("Cannot change cloned service\n");)
			error = EPERM;
			break;
		}
		if(conf->options & LBDR_OPTIONS_SERVICE_TIMEOUT) {
			if(conf->options & LBDR_OPTIONS_TIMEOUT_SYN)
				service->to->t_syn = conf->t_syn * nticks;
			if(conf->options & LBDR_OPTIONS_TIMEOUT_EST)
				service->to->t_est = conf->t_est * nticks;
			if(conf->options & LBDR_OPTIONS_TIMEOUT_CLO)
				service->to->t_clo = conf->t_clo * nticks;
			D(uprintf("drlb: Set timeout for %s: Syn: %i\tEstablish: %i\tClosed: %i\n", paddr(service->ip), service->to->t_syn/nticks, service->to->t_est/nticks, service->to->t_clo/nticks);)
		}
		oldsrv = service->server;
		newsrv = add_server(oldsrv, service->nb_server, &(conf->ip_srv), conf->weight, conf->rate, &error);
		if(newsrv) {
			service->server = newsrv;
			service->nb_server++;
			FOREACHCLONE(service,
				clone->server = service->server;
				clone->nb_server = service->nb_server
			)
			if(oldsrv) free(oldsrv, M_BUF);
		}
                break;
        case LBDRDELCONF:
		DG(log(LOG_DEBUG, "LBDRDELCONF\n"));
		service = find_service(dst);
                if(!service) {
			error = ENOENT;
			break;
                }
		if(service->cloned) {
			D(uprintf("Cannot change cloned service\n");)
			error = EPERM;
			break;
		}
		if(is_any(&(conf->ip_srv))) {
			if(conf->force) {
				unregister_service(service->ip);
				free_service((void *)service);
			} else error = EACCES;
			break;
		}
		oldsrv = service->server;
		newsrv = delete_server(oldsrv, &(conf->ip_srv), service->nb_server, &error);
		if(newsrv) {
			if(service->nb_server == 1) {
				unregister_service(dst);
				free(newsrv, M_BUF);
				free_service((void *)service);
			} else {
				service->nb_server--;
				service->used_server %= service->nb_server;
				service->server = newsrv;
				FOREACHCLONE(service,
					clone->nb_server = service->nb_server;
					clone->used_server %= clone->nb_server;
					clone->server = service->server;
				)
				free(oldsrv, M_BUF);
			}
                }
                break;
        case LBDRCHGCONF:
        	DG(log(LOG_DEBUG, "LBDRCHGCONF\n");)
        	service = find_service(dst);
        	if(!service) {
			error = ENOENT;
        		break;
        	}
		if(service->cloned) {
			D(uprintf("Cannot change cloned service\n");)
			error = EPERM;
			break;
		}
		if(conf->options & LBDR_OPTIONS_SERVICE_TIMEOUT) {
			if(conf->options & LBDR_OPTIONS_TIMEOUT_SYN)
				service->to->t_syn = conf->t_syn * nticks;
			if(conf->options & LBDR_OPTIONS_TIMEOUT_EST)
				service->to->t_est = conf->t_est * nticks;
			if(conf->options & LBDR_OPTIONS_TIMEOUT_CLO)
				service->to->t_clo = conf->t_clo * nticks;
			D(uprintf("drlb: New timeout for %s: Syn: %i\tEstablish: %i\tClosed: %i\n", paddr(service->ip), service->to->t_syn/nticks, service->to->t_est/nticks, service->to->t_clo/nticks);)
		}
		if(conf->options & LBDR_OPTIONS_SERVICE_PERSIST) {
			D(if(service->persist != conf->persist)
				uprintf("Switching %s to %sSSL\n", paddr(service->ip), conf->persist ? "" : "no-");)
			service->persist = conf->persist;
		}
		if(conf->options & LBDR_OPTIONS_SERVICE_SCHED) {
			switch(conf->sched) {
			case LBDR_SCHED_WLC:
				D(uprintf("Switching to Weighted-Least-Connections scheduler\n");)
				service->sched = scheduler_wlc;
				break;
			case LBDR_SCHED_WRR:
				D(uprintf("Swtching to Weighted-Round-Robin scheduler\n");)
				service->sched = scheduler_wrr;
				break;
			case LBDR_SCHED_LC:
				D(uprintf("Swtching to Least-Connections scheduler\n");)
				service->sched = scheduler_lc;
				break;
			case LBDR_SCHED_RR:
				D(uprintf("Switching to Round-Robin scheduler\n");)
				service->sched = scheduler_rr;
				break;
			default:
				error = EINVAL;
				break;
			}
		}
		if(conf->options & LBDR_OPTIONS_SERVER) {

			error = modify_server(service->server, &(conf->ip_srv), service->nb_server, conf->options, conf->weight, conf->rate, conf->status);
		}
		FOREACHCLONE(service,
			clone->sched = service->sched;
			clone->persist = service->persist;
		)
        	break;
        /*
        case LBDRCHGSRV:
		DG(log(LOG_DEBUG, "LBDRCHGSRV\n");)
		service = find_service(dst);
        	if(!service) {
			error = ENOENT;
			break;
		}
		error = modify_server(service->server, &(conf->ip_srv), service->nb_server, conf->weight, conf->rate, conf->status);
        	break;
	*/
        case LBDRCLONE:
        	DG(log(LOG_DEBUG, "LBDRCLONE\n");)
		service = find_service(dst);
        	if(!service) {
        		error = ENOENT;
        		break;
        	}
		if(service->cloned) {
			D(uprintf("Cannot clone already cloned service\n");)
			error = EPERM;
			break;
		}
		clone_service(service, &(conf->ip_srv));
        	break;
        case LBDRINFO:
		{
			struct lbdr_info *inf = (struct lbdr_info *)arg;
			if(brother.ip->ss_family) {
				copy_addr(&(inf->brother_ip), brother.ip);
				inf->brother_key = sync_key;
			}
			inf->nb_cnx = 0;
			inf->registred_ip = get_nb_ip(&(inf->nb_cnx));
			inf->err_nomem = stats_nomem;
			inf->err_notsame = stats_notsame;
			inf->err_dropact = stats_dropact;
			inf->err_nombuf = stats_nombuf;
			inf->err_badmbuf = stats_badmbuf;
			inf->err_udpbadcsum = stats_udpbadcsum;
			inf->err_udpbadlen = stats_udpbadlen;
		}
        	break;
        case LBDRDEBUG:
			dump_time();
        	break;
        case LBDRSTATS:
                error = dump_stats((struct lbdr_param *)arg);
		break;
	case LBDRCLEAR:
		DG(log(LOG_DEBUG, "LBDRCHGCLEAR\n");)
		stats_nomem = 0;
		stats_badmbuf = 0;
		stats_nombuf = 0;
		stats_dropact = 0;
#ifdef DEBUGTIME
		init_timelost();
#endif
		i = 0;
		while(i < LIMIT_SERVICE) {
			service = service_available[i];
			while(service) {
				last = service;
				service = service->next;
				if(!last->cloned) {
					unregister_service(last->ip);
					free_service((void *)last);
				}
			}
			i++;
		}
		bzero(service_available, sizeof(struct lbdr_service) * LIMIT_SERVICE);
                break;

        case LBDRSETSHARE:
        	{
			struct lbdr_share *shr = (struct lbdr_share *)arg;
			DG(log(LOG_DEBUG, "LBDRSETSHARE\n");)

			sync_key = shr->key;
			copy_addr(brother.ip, &(shr->ip_dst));

			brother.rt = rtalloc1((struct sockaddr *)(brother.ip), 1, 0UL);

			if(brother.rt) {
				/* FIXME: rtfree ? */
				/*N(brother.rt->rt_refcnt--;) */
				F(RT_UNLOCK(brother.rt);)
				D(uprintf("Sharing connections with %s.%i via %s with key: 0x%x\n", paddr(brother.ip), ntohs(SPORT(brother.ip)), brother.rt->rt_ifp->if_xname, sync_key));
			} else {
				log(LOG_ERR, "drlb: Error: Host %s is unreachable!\n", paddr(brother.ip));
			}
		}

        	break;

        default:
                D(log(LOG_DEBUG, "Unknow ioctl received\n");)
                error = EINVAL;
		break;
	}

        return (error);
}

/***
 * Module functions
 *
 * (load, unload, init, deinit)
 *
 */

#ifdef __FreeBSD__

static int drlb_load(struct module *m, int what, void *arg)
{
	struct pfil_head *pfh_inet4, *pfh_inet6;
	int err = 0;

	pfh_inet4 = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	pfh_inet6 = pfil_head_get(PFIL_TYPE_AF, AF_INET6);
	if (!pfh_inet4 || !pfh_inet6) return ESRCH;

	switch(what) {
	case MOD_LOAD:

		err = init_module();
		if(err) return err;
		pfil_add_hook(lb4_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet4);
		pfil_add_hook(lb6_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet6);

		break;


	case MOD_UNLOAD:

		pfil_remove_hook(lb4_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet4);
		pfil_remove_hook(lb6_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet6);
		err = deinit_module(); /* FIXME: and what if error ? */

		break;

	case MOD_QUIESCE:
		break;  /* FIXME: return EBUSY if in use */
	case MOD_SHUTDOWN :
		pfil_remove_hook(lb4_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet4);
		pfil_remove_hook(lb6_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet6);
		break;

	default:
		err = EINVAL;
		break;
	}

	return err;
}


static moduledata_t drlb_mod = { "drlb", drlb_load, NULL };
DECLARE_MODULE(drlb, drlb_mod, SI_SUB_KLD, SI_ORDER_ANY);
MODULE_VERSION(drlb, 1);


#endif




#ifdef __NetBSD__


static int drlbopen(dev_t device, int flags, int fmt, struct lwp *process)
{
	return 0;
}

static int drlbclose(dev_t device, int flags, int fmt, struct lwp *process)
{
	return 0;
}

const struct cdevsw drlb_cdevsw = {
	drlbopen, drlbclose, noread, nowrite, drlbioctl,
	nostop, notty, nopoll, nommap, nokqfilter, D_OTHER,
};



/* TODO: manage error */
void drlbattach(int num)
{
	int cmajor = -1, bmajor = -1;

	pfh_inet4 = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	pfh_inet6 = pfil_head_get(PFIL_TYPE_AF, AF_INET6);

	IF_BIG(mtx_init(&big_lock, "mtxbig", NULL, MTX_DEF);)

	/*nticks = mstohz(1000);*/
	nticks = hz;

        malloc_type_attach(M_BUF);

	devsw_attach("drlb", NULL, &bmajor, &drlb_cdevsw, &cmajor);

	service_available = malloc(sizeof(struct lbdr_service) * LIMIT_SERVICE, M_BUF, M_NOWAIT | M_ZERO);
        bzero(&brother, sizeof(struct lbdr_server));
	brother.ip = malloc(sizeof(struct sockaddr_storage), M_BUF, M_NOWAIT | M_ZERO);

	if(!service_available || !brother.ip) { 
		log(LOG_ERR, "drlb: not enough kernel memory!\n");
		return;
	}

	/*
	stats_nomem = 0;
	stats_badmbuf = 0;
	stats_nombuf = 0;
	stats_dropact = 0;
	sync_port = htons(OUT_UDP_PORT);
	*/

#ifdef DEBUGTIME
	init_timelost();
#endif

	if (pfh_inet4) {
		pfil_add_hook(lb4_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet4);
	}

	if(pfh_inet6) {
		pfil_add_hook(lb6_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet6);
	}

	log(LOG_INFO, "drlb: <HS: %i  %iHz>\n", LIMIT_SERVICE, nticks);
}

void drlbdetach(int num)
{
	struct lbdr_service *tmp, *last;
	int i = LIMIT_SERVICE - 1;

	if(pfh_inet4) {
		pfil_remove_hook(lb4_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet4);
	}

	if(pfh_inet6) {
		pfil_remove_hook(lb6_input, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet6);
	}

	while(i) {
		tmp = service_available[i];
		while(tmp) {
			last = tmp;
			tmp = tmp->next;
			if(!last->cloned) {
				unregister_service(last->ip);
				free_service((void *)last);
			}
		}
		i--;
	}

	devsw_detach(NULL, &drlb_cdevsw);

#ifdef DEBUGTIME
	clear_timelost();
#endif

	free(brother.ip, M_BUF);
	free(service_available, M_BUF);
        malloc_type_detach(M_BUF);

	IF_BIG(mtx_destroy(&big_lock);)

}


#ifdef _MODULE

static int kld_loaded = 0;

#include <sys/module.h>

MODULE(MODULE_CLASS_MISC, drlb, NULL);

static int drlb_modcmd(modcmd_t what, void *arg)
{
	switch(what) {

	case MODULE_CMD_INIT:

		if (kld_loaded) return EEXIST;

		drlbattach(0);

		kld_loaded = 1;

		/*log(LOG_INFO, "drlb: inserted (pfil)" D(" with debug") DG(", verbose") DTR(", trace") "\n");*/
		uprintf("drlb kld version %s. %iHz. HSservice: %i" D(" with debug") DG(", verbose") DTR(", trace") "\n", LBDR_VERSION, nticks, LIMIT_SERVICE);

#ifndef NOBIG
		log(LOG_INFO, "drlb: [BIG-LOCKED]\n");
#endif

		break;

	case MODULE_CMD_FINI:

		if(!kld_loaded) return 0;
		kld_loaded = 0;
		drlbdetach(0);
		uprintf("drlb: unloaded\n");

		break;

	case MODULE_CMD_STAT:
		break;
	default:
		return EINVAL;
		break;
	}

	return 0;
}


#endif /* MODULE */


#endif /* NetBSD*/




/* EOF */
