/*-
 * Copyright (c) 2008 Jean-Yves Moulin <jym@ldh.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#ifndef __LB_DR_H__
#define __LB_DR_H__


/* le device a ouvrir pour configurer le LB: */
#define LBDR_DEVICE "/dev/lbdr"

/* les differents ordonnanceurs: */
#define LBDR_SCHED_RR	0x01
#define LBDR_SCHED_WRR	0x02
#define LBDR_SCHED_LC   0x04
#define LBDR_SCHED_WLC  0x08



#define LBDR_OPTIONS_SERVICE		0x00f
#define LBDR_OPTIONS_SERVICE_SCHED	0x001
#define LBDR_OPTIONS_SERVICE_TIMEOUT	0x002
#define LBDR_OPTIONS_SERVICE_PERSIST	0x004
#define LBDR_OPTIONS_SERVER		0x0f0
#define LBDR_OPTIONS_SERVER_WEIGHT	0x010
#define LBDR_OPTIONS_SERVER_RATE	0x020
#define LBDR_OPTIONS_SERVER_STATUS	0x040
#define LBDR_OPTIONS_TIMEOUT		0xf00
#define LBDR_OPTIONS_TIMEOUT_SYN	0x100
#define LBDR_OPTIONS_TIMEOUT_EST	0x200
#define LBDR_OPTIONS_TIMEOUT_CLO	0x400

/* la structure pour configurer le LB: */
struct lbdr_conf {
	struct sockaddr_storage ip_dst;
	struct sockaddr_storage ip_srv;
	uint32_t options;
	uint32_t hashtablesize;
	uint32_t t_syn;
	uint32_t t_est;
	uint32_t t_clo;
	uint32_t rate;
	uint32_t weight:15,
		 status:2,
		 sched:4,
		 persist:1,
		 force:1,
		 clone:1;
};

struct lbdr_stats_service {
	unsigned long ls_coll;
	uint32_t ls_timeout_syn;
	uint32_t ls_timeout_est;
	uint32_t ls_timeout_clo;
	uint32_t ls_err_nosyn;
	uint32_t ls_err_nosrv;
	uint32_t ls_hashtablesize:26,
		 ls_sched:4,
		 ls_persist:1;
};

struct lbdr_stats_server {
	unsigned long ls_nb_cnx;
	unsigned long ls_pkts;
	unsigned long ls_bytes;
	unsigned long ls_no_fin;
	uint32_t ls_act_cnx;
	uint32_t ls_inact_cnx;
	uint32_t ls_shared_cnx;
	uint32_t ls_err_unreach;
	uint32_t ls_err_noroute;
	uint32_t ls_weight:10,
		 ls_status:1,
		 ls_rate:16;
};

/* for the type field: */
#define LBDR_TYPE_SERVICE	0x01
#define LBDR_TYPE_CLONE		0x02
#define LBDR_TYPE_SERVER	0x04

struct lbdr_stats {
	uint32_t type; /* id + status (clone, service, server..) */
	struct sockaddr_storage ip;
	union {
		struct lbdr_stats_service svi;
		struct lbdr_stats_server  srv;
	} stats;
};


struct lbdr_info {
	struct sockaddr_storage brother_ip;
	uint32_t brother_key;
	uint32_t registred_ip;
	unsigned long nb_cnx;
	unsigned long err_nomem;
	unsigned long err_notsame;
	unsigned long err_dropact;
	unsigned long err_nombuf;
	unsigned long err_badmbuf;
	unsigned long err_udpbadcsum;
	unsigned long err_udpbadlen;
};


struct lbdr_share {
	struct sockaddr_storage ip_dst;
	uint32_t key;
};


/* pour recuperer les statistiques: */
struct lbdr_param {
	void *data;
	size_t size;
};

/* Les definitions des ioctl: */
#define		LBDRADDCONF	_IOW('L', 1, struct lbdr_conf)
#define		LBDRDELCONF	_IOW('L', 2, struct lbdr_conf)
#define		LBDRCHGCONF	_IOW('L', 3, struct lbdr_conf)
#define		LBDRCLONE	_IOW('L', 10, struct lbdr_conf)

/*
#define		LBDRCHGSVICE	_IOW('L', 4, struct lbdr_conf)
#define		LBDRCHGTO	_IOW('L', 5, struct lbdr_timeout)
*/

#define		LBDRCLEAR	_IOW('L', 6, int)

#define		LBDRINFO	_IOR('L', 7, struct lbdr_info)
#define		LBDRDEBUG	_IOR('L', 12, int)

#define		LBDRSTATS	_IOWR('L', 8, struct lbdr_param)

#define		LBDRSETSHARE	_IOW('L', 9, struct lbdr_share)
#define		LBDRGETSHARE	_IOR('L', 11, struct lbdr_share)


#endif
